"""Test of filter functionalities"""
import pathlib

from shapeout2.gui.main import ShapeOut2
from shapeout2.gui.analysis import DlgSlotReorder
from shapeout2 import session
import pytest

datapath = pathlib.Path(__file__).parent / "data"


@pytest.fixture(autouse=True)
def run_around_tests():
    # Code that will run before your test, for example:
    session.clear_session()
    # A test function will be run at this point
    yield
    # Code that will run after your test, for example:
    session.clear_session()


def test_filter_min_max_inf(qtbot):
    mw = ShapeOut2()
    qtbot.addWidget(mw)

    # add 3 dataslots
    path = datapath / "calibration_beads_47.rtdc"
    mw.add_dataslot(paths=[path, path, path])

    # change the slot names
    for ii, slot in enumerate(mw.pipeline.slots):
        slot.name = "slot numero {}".format(ii)
        mw.adopt_slot(slot.__getstate__())

    # sanity checks
    assert mw.pipeline.slots[0].name == "slot numero 0"
    assert mw.pipeline.slots[1].name == "slot numero 1"
    assert mw.pipeline.slots[2].name == "slot numero 2"

    # create reorder dialog manually
    dlg = DlgSlotReorder(mw.pipeline, mw)
    dlg.pipeline_changed.connect(mw.adopt_pipeline)
    # reorder plots
    dlg.listWidget.setCurrentRow(0)
    dlg.toolButton_down.clicked.emit()
    dlg.on_ok()

    # now check that reordering happened
    assert mw.pipeline.slots[0].name == "slot numero 1"
    assert mw.pipeline.slots[1].name == "slot numero 0"
    assert mw.pipeline.slots[2].name == "slot numero 2"



"""Facilitate caching of plot data"""
from . import util


def get_contour_data(rtdc_ds, xax, yax, xacc, yacc, xscale, yscale,
                     kde_type="histogram", kde_kwargs={}):
    rtdc_ds.apply_filter()
    cfg = rtdc_ds.config
    tohash = [
        rtdc_ds.identifier, rtdc_ds.filter.all,
        cfg["calculation"] if "calculation" in cfg else "",
        xax, yax, xacc, yacc, xscale, yscale,
        kde_type, kde_kwargs]
    shash = util.hashobj(tohash)
    if shash in cache_data:
        x, y, den = cache_data[shash]
    else:
        # compute scatter plot data
        x, y, den = rtdc_ds.get_kde_contour(
            xax=xax,
            yax=yax,
            xacc=xacc,
            yacc=yacc,
            xscale=xscale,
            yscale=yscale,
            kde_type=kde_type,
            kde_kwargs=kde_kwargs)
        # save in cache
        cache_data[shash] = x, y, den
    return x, y, den


def get_scatter_data(rtdc_ds, downsample, xax, yax, xscale, yscale,
                     kde_type="histogram", kde_kwargs={}):
    rtdc_ds.apply_filter()
    cfg = rtdc_ds.config
    tohash = [
        rtdc_ds.identifier, rtdc_ds.filter.all, downsample,
        cfg["calculation"] if "calculation" in cfg else "",
        xax, yax, xscale, yscale, kde_type, kde_kwargs]
    shash = util.hashobj(tohash)
    if shash in cache_data:
        x, y, kde, idx = cache_data[shash]
    else:
        # compute scatter plot data
        x, y, idx = rtdc_ds.get_downsampled_scatter(
            xax=xax,
            yax=yax,
            downsample=downsample,
            xscale=xscale,
            yscale=yscale,
            remove_invalid=True,
            ret_mask=True)
        # kde
        kde = rtdc_ds.get_kde_scatter(
            xax=xax,
            yax=yax,
            positions=(x, y),
            kde_type=kde_type,
            kde_kwargs=kde_kwargs,
            xscale=xscale,
            yscale=yscale)
        if kde.size and kde.min() != kde.max():
            kde -= kde.min()
            kde /= kde.max()
        # save in cache
        cache_data[shash] = x, y, kde, idx
    return x, y, kde, idx


cache_data = {}



import copy
import pathlib
import tempfile

import dclab
import numpy as np
from shapeout2 import pipeline


def test_apply_filter_ray():
    path = pathlib.Path(__file__).parent / "data" / "calibration_beads_47.rtdc"

    # initiate the pipeline
    pl = pipeline.Pipeline()
    slot_id = pl.add_slot(path=path)

    # add a filter
    filt_id1 = pl.add_filter()
    filt1 = pl.get_filter(filt_id1)
    amin, amax = pl.get_min_max("area_um")
    filt1.boxdict["area_um"] = {"start": amin,
                                "end": (amin + amax)/2,
                                "active": True}
    pl.set_element_active(slot_id, filt_id1)

    # and another one
    filt_id2 = pl.add_filter()
    filt2 = pl.get_filter(filt_id2)
    dmin, dmax = pl.get_min_max("deform")
    filt2.boxdict["deform"] = {"start": dmin,
                               "end": (dmin + dmax)/2,
                               "active": True}
    pl.set_element_active(slot_id, filt_id2)

    ds_ref = pl.get_dataset(0)  # this will run through the filter ray
    ds0 = dclab.new_dataset(path)
    ds_ext = pl.apply_filter_ray(ds0, slot_id)  # this should do the same thing

    for feat in ds_ref.features_loaded:
        if dclab.dfn.scalar_feature_exists(feat):
            assert np.allclose(ds_ref[feat], ds_ext[feat]), feat
    assert np.all(ds_ref.filter.all == ds_ext.filter.all)


def test_get_min_max_inf():
    # generate fake dataset
    path = pathlib.Path(__file__).parent / "data" / "calibration_beads_47.rtdc"
    with dclab.new_dataset(path) as ds:
        config = copy.deepcopy(ds.config)

    tmp = tempfile.mktemp(".rtdc", prefix="example_filter_inf_")
    ds2 = dclab.new_dataset({"deform": np.linspace(0, .01, 100),
                             "area_um": np.linspace(20, 200, 100),
                             "area_ratio": np.linspace(1, 1.1, 100)
                             })
    ds2.config.update(config)
    ds2["area_ratio"][0] = np.inf
    ds2["area_ratio"][1] = np.nan
    ds2.export.hdf5(tmp, features=["area_um", "deform", "area_ratio"])

    # initiate the pipeline
    pl = pipeline.Pipeline()
    pl.add_slot(path=tmp)
    pl.add_filter()

    # get the current min/max values
    amin, amax = pl.get_min_max("area_ratio")
    assert amin == ds2["area_ratio"][2]
    assert amax == 1.1

    try:
        pathlib.Path(tmp).unlink()
    except BaseException:
        pass


def test_get_min_max_plot():
    """See #22"""
    path = pathlib.Path(__file__).parent / "data" / "calibration_beads_47.rtdc"

    # initiate the pipeline
    pl = pipeline.Pipeline()
    slot_id = pl.add_slot(path=path)

    # add a filter
    filt_id = pl.add_filter()
    filt = pl.get_filter(filt_id)

    # get the current min/max values
    amin, amax = pl.get_min_max("area_um")

    # modify the filter
    filt.boxdict["area_um"] = {"start": amin,
                               "end": (amin + amax)/2,
                               "active": True}

    # make the filter active in the filter ray
    pl.set_element_active(slot_id, filt_id)

    # add a plot
    plot_id = pl.add_plot()
    pl.set_element_active(slot_id, plot_id)

    # get the new min/max values
    amin2, amax2 = pl.get_min_max("area_um", plot_id=plot_id)

    assert amin == amin2
    assert amax2 <= (amin + amax) / 2


if __name__ == "__main__":
    # Run all tests
    loc = locals()
    for key in list(loc.keys()):
        if key.startswith("test_") and hasattr(loc[key], "__call__"):
            loc[key]()



import os
import pathlib
import pkg_resources
import signal
import sys
import traceback
import webbrowser

import dclab
import h5py
import numpy
import scipy

from PyQt5 import uic, QtCore, QtGui, QtWidgets
import pyqtgraph as pg

from . import analysis
from . import bulk
from . import compute
from . import dcor
from . import export
from . import pipeline_plot
from . import preferences
from . import quick_view
from . import update
from . import widgets

from .. import pipeline
from .. import session


from .._version import version as __version__


# global plotting configuration parameters
pg.setConfigOption("background", "w")
pg.setConfigOption("foreground", "k")
pg.setConfigOption("antialias", True)
pg.setConfigOption("imageAxisOrder", "row-major")


# set Qt icon theme search path
QtGui.QIcon.setThemeSearchPaths([
    os.path.join(pkg_resources.resource_filename("shapeout2", "img"),
                 "icon-theme")])
QtGui.QIcon.setThemeName(".")


class ShapeOut2(QtWidgets.QMainWindow):
    plots_changed = QtCore.pyqtSignal()

    def __init__(self):
        """Initialize Shape-Out 2

        If you pass the "--version" command line argument, the
        application will print the version after initialization
        and exit.
        """
        QtWidgets.QMainWindow.__init__(self)
        path_ui = pkg_resources.resource_filename("shapeout2.gui", "main.ui")
        uic.loadUi(path_ui, self)
        # update check
        self._update_thread = None
        self._update_worker = None
        # Settings are stored in the .ini file format. Even though
        # `self.settings` may return integer/bool in the same session,
        # in the next session, it will reliably return strings. Lists
        # of strings (comma-separated) work nicely though.
        QtCore.QCoreApplication.setOrganizationName("Zellmechanik-Dresden")
        QtCore.QCoreApplication.setOrganizationDomain("zellmechanik.com")
        QtCore.QCoreApplication.setApplicationName("shapeout2")
        QtCore.QSettings.setDefaultFormat(QtCore.QSettings.IniFormat)
        #: Shape-Out settings
        self.settings = QtCore.QSettings()
        self.settings.setIniCodec("utf-8")
        # Register user-defined DCOR API Key in case the user wants to
        # open a session with private data.
        api_key = self.settings.value("dcor/api key", "")
        dclab.rtdc_dataset.fmt_dcor.APIHandler.add_api_key(api_key)
        #: Analysis pipeline
        self.pipeline = pipeline.Pipeline()
        # GUI
        self.setWindowTitle("Shape-Out {}".format(__version__))
        # Disable native menu bar (e.g. on Mac)
        self.menubar.setNativeMenuBar(False)
        # File menu
        self.actionLoadDataset.triggered.connect(self.add_dataslot)
        self.actionLoadDCOR.triggered.connect(self.on_action_dcor)
        self.actionClearSession.triggered.connect(self.on_action_clear)
        self.actionOpenSession.triggered.connect(self.on_action_open)
        self.actionQuit.triggered.connect(self.on_action_quit)
        self.actionSaveSession.triggered.connect(self.on_action_save)
        # Edit menu
        self.actionChangeDatasetOrder.triggered.connect(
            self.on_action_change_dataset_order)
        self.actionPreferences.triggered.connect(self.on_action_preferences)
        # Compute menu
        self.actionComputeStatistics.triggered.connect(
            self.on_action_compute_statistics)
        self.actionComputeEmodulus.triggered.connect(
            self.on_action_compute_emodulus)
        # Export menu
        self.actionExportData.triggered.connect(self.on_action_export_data)
        self.actionExportFilter.triggered.connect(self.on_action_export_filter)
        self.actionExportPlot.triggered.connect(self.on_action_export_plot)
        # Import menu
        self.actionImportFilter.triggered.connect(self.on_action_import_filter)
        # Help menu
        self.actionDocumentation.triggered.connect(self.on_action_docs)
        self.actionSoftware.triggered.connect(self.on_action_software)
        self.actionAbout.triggered.connect(self.on_action_about)
        # Subwindows
        self.subwindows = {}
        # Subwindows for plots
        self.subwindows_plots = {}
        # Initialize a few things
        self.init_quick_view()
        self.init_analysis_view()
        self.mdiArea.cascadeSubWindows()
        # BLOCK MATRIX (wraps DataMatrix and PlotMatrix)
        # BlockMatrix appearance
        self.toolButton_dm.clicked.connect(self.on_data_matrix)
        self.splitter.splitterMoved.connect(self.on_splitter)
        # BlockMatrix Actions
        self.actionNewFilter.triggered.connect(self.add_filter)
        self.actionNewPlot.triggered.connect(self.add_plot)
        self.block_matrix.toolButton_load_dataset.clicked.connect(
            self.add_dataslot)
        self.block_matrix.toolButton_new_filter.clicked.connect(
            self.add_filter)
        self.block_matrix.toolButton_new_plot.clicked.connect(self.add_plot)
        # BlockMatrix default state
        self.toolButton_new_plot.setEnabled(False)
        self.block_matrix.toolButton_new_plot.setEnabled(False)
        # BlockMatrix other signals
        self.block_matrix.pipeline_changed.connect(self.adopt_pipeline)
        self.block_matrix.slot_modify_clicked.connect(self.on_modify_slot)
        self.block_matrix.filter_modify_clicked.connect(self.on_modify_filter)
        self.block_matrix.plot_modify_clicked.connect(self.on_modify_plot)
        # ANALYSIS VIEW
        self.widget_ana_view.set_pipeline(self.pipeline)
        # filter signals
        self.widget_ana_view.filter_changed.connect(self.adopt_filter)
        self.widget_ana_view.pipeline_changed.connect(self.adopt_pipeline)
        # polygon filter creation
        self.widget_ana_view.widget_filter.request_new_polygon_filter.connect(
            self.on_new_polygon_filter)
        self.widget_quick_view.polygon_filter_created.connect(
            self.widget_ana_view.widget_filter.update_polygon_filters)
        self.widget_quick_view.polygon_filter_modified.connect(
            self.widget_ana_view.widget_filter.update_polygon_filters)
        self.widget_quick_view.polygon_filter_modified.connect(
            self.on_quickview_refresh)  # might be an active filter (#26)
        self.widget_quick_view.polygon_filter_modified.connect(
            self.plots_changed)  # might be an active filter (#26)
        # This is important, because if meta data such as emodulus recipe
        # is changed, the QuickView must be updated as well.
        self.plots_changed.connect(self.widget_quick_view.plot)
        # plot signals
        self.widget_ana_view.plot_changed.connect(self.adopt_plot)
        # slot signals
        self.widget_ana_view.slot_changed.connect(self.adopt_slot)
        # if "--version" was specified, print the version and exit
        if "--version" in sys.argv:
            print(__version__)
            QtWidgets.QApplication.processEvents()
            sys.exit(0)
        # check pyqtgraph version
        self._check_pg_version()
        # check for updates
        do_update = int(self.settings.value("check for updates", 1))
        self.on_action_check_update(do_update)
        # finalize
        self.show()
        self.raise_()
        self.activateWindow()
        self.showMaximized()
        self.setWindowState(QtCore.Qt.WindowState.WindowActive)

    def _check_pg_version(self):
        """Tells the user if the pyqtgraph version is not correct"""
        if (int(self.settings.value("advanced/check pyqtgraph version", 1))
                and pg.__version__ != "0.11.0.dev0+g1c63ae5"):
            msg = QtWidgets.QMessageBox()
            msg.setIcon(QtWidgets.QMessageBox.Warning)
            msg.setText("You are using an unexpected version of pyqtgraph. "
                        + "Plot export may not work correctly!")
            msg.setWindowTitle("pyqtgraph version mismatch")
            msg.setDetailedText(
                "If you have installed Shape-Out using pip, please "
                + "install our pyqtgraph fork using: \n\n"
                + "pip install "
                + "git+git://github.com/ZELLMECHANIK-DRESDEN/pyqtgraph.git"
            )
            msg.exec_()

    @widgets.show_wait_cursor
    @QtCore.pyqtSlot(dict)
    def adopt_filter(self, filt_state):
        filt_id = filt_state["identifier"]
        state = self.pipeline.__getstate__()
        for ii in range(len(state["filters"])):
            if state["filters"][ii]["identifier"] == filt_id:
                state["filters"][ii] = filt_state
                # make sure filters are enabled/disabled
                if (filt_state["filter used"]
                        and filt_id not in state["filters used"]):
                    state["filters used"].append(filt_id)
                elif (not filt_state["filter used"]
                      and filt_id in state["filters used"]):
                    state["filters used"].remove(filt_id)
                break
        else:
            raise ValueError("Filter not in pipeline: {}".format(filt_id))
        self.adopt_pipeline(state)

    @widgets.show_wait_cursor
    @QtCore.pyqtSlot(dict)
    def adopt_pipeline(self, pipeline_state):
        # If the number of subplots within a plot changed, update the
        # plot size accordingly.
        for plot_index, plot_id in enumerate(self.pipeline.plot_ids):
            old_ncol, old_nrow = self.pipeline.get_plot_col_row_count(plot_id)
            try:
                new_ncol, new_nrow = self.pipeline.get_plot_col_row_count(
                    plot_id, pipeline_state)
                lay = pipeline_state["plots"][plot_index]["layout"]
            except (KeyError, IndexError):
                # the plot was removed
                continue
            else:
                lay["size x"] += 200*(new_ncol-old_ncol)
                lay["size y"] += 200*(new_nrow-old_nrow)
        # set the new state of the pipeline
        self.pipeline.__setstate__(pipeline_state)
        # update BlockMatrix
        if self.sender() != self.block_matrix:
            # Update BlockMatrix
            self.block_matrix.adopt_pipeline(pipeline_state)
        # trigger redraw
        self.block_matrix.update()
        # Invalidate block matrix elements that do not make sense due to
        # filtering or plotting features.
        invalid_dm = []
        invalid_pm = []
        for slot_index, slot in enumerate(self.pipeline.slots):
            ds = self.pipeline.get_dataset(slot_index=slot_index,
                                           filt_index=None)
            for filt in self.pipeline.filters:
                # box filters
                for feat in filt.boxdict:
                    if feat not in ds.features:
                        invalid_dm.append((slot.identifier, filt.identifier))
                        break
                else:
                    # polygon filters
                    for pid in filt.polylist:
                        pf = dclab.PolygonFilter.get_instance_from_id(pid)
                        if (pf.axes[0] not in ds.features
                                or pf.axes[1] not in ds.features):
                            invalid_dm.append((slot.identifier,
                                               filt.identifier))
                            break
            for plot in self.pipeline.plots:
                plot_state = plot.__getstate__()
                if (plot_state["general"]["axis x"] not in ds
                        or plot_state["general"]["axis x"] not in ds):
                    invalid_pm.append((slot.identifier, plot.identifier))
        self.block_matrix.invalidate_elements(invalid_dm, invalid_pm)
        # Update AnalysisView
        self.widget_ana_view.adopt_pipeline(pipeline_state)
        # Update QuickView choices
        self.widget_quick_view.update_feature_choices()
        # update list of polygon filters in Quick View
        self.widget_quick_view.update_polygon_panel()
        # Show Plot Windows
        # create and show
        for plot_state in pipeline_state["plots"]:
            self.add_plot_window(plot_state["identifier"])
        # remove old plot subwindows
        plot_ids = [pp["identifier"] for pp in pipeline_state["plots"]]
        for plot_id in list(self.subwindows_plots.keys()):
            if plot_id not in plot_ids:
                sub = self.subwindows_plots.pop(plot_id)
                for child in sub.children():
                    # disconnect signals
                    if isinstance(child, pipeline_plot.PipelinePlot):
                        self.plots_changed.disconnect(child.update_content)
                        break
                sub.deleteLater()
        self.plots_changed.emit()
        self.widget_ana_view.widget_plot.update_content()
        # Remove zombie slots
        for slot_id in list(pipeline.Dataslot._instances.keys()):
            if slot_id not in self.pipeline.slot_ids:
                pipeline.Dataslot.remove_slot(slot_id)
        # enable buttons
        if self.pipeline.slots:
            self.toolButton_new_plot.setEnabled(True)
            self.block_matrix.toolButton_new_plot.setEnabled(True)
        else:
            self.toolButton_new_plot.setEnabled(False)
            self.block_matrix.toolButton_new_plot.setEnabled(False)
        # redraw
        self.block_matrix.update()
        self.mdiArea.update()
        self.subwindows["analysis_view"].update()

    @widgets.show_wait_cursor
    @QtCore.pyqtSlot(dict)
    def adopt_plot(self, plot_state):
        plot_id = plot_state["identifier"]
        state = self.pipeline.__getstate__()
        for ii in range(len(state["plots"])):
            if state["plots"][ii]["identifier"] == plot_id:
                state["plots"][ii] = plot_state
                break
        else:
            raise ValueError("Plot not in pipeline: {}".format(plot_id))
        self.adopt_pipeline(state)

    @widgets.show_wait_cursor
    @QtCore.pyqtSlot(dict)
    def adopt_slot(self, slot_state):
        slot_id = slot_state["identifier"]
        state = self.pipeline.__getstate__()
        for ii in range(len(state["slots"])):
            if state["slots"][ii]["identifier"] == slot_id:
                state["slots"][ii] = slot_state
                # make sure filters are enabled/disabled
                if (slot_state["slot used"]
                        and slot_id not in state["slots used"]):
                    state["slots used"].append(slot_id)
                elif (not slot_state["slot used"]
                      and slot_id in state["slots used"]):
                    state["slots used"].remove(slot_id)
                break
        else:
            raise ValueError("Slot not in pipeline: {}".format(slot_id))
        self.adopt_pipeline(state)

    @widgets.show_wait_cursor
    @QtCore.pyqtSlot()
    def add_dataslot(self, paths=None, is_dcor=False):
        """Adds a dataslot to the pipeline

        Parameters
        ----------
        paths: list of str or None
            If specified, no file dialog is displayed and the files
            specified are loaded. Can also be DCOR URL.
        is_dcor: bool
            If set to True, `paths` will be treated as a list of
            DCOR URLs. Does not have any effect if `paths` is None.
        """
        if paths is None:
            fnames, _ = QtWidgets.QFileDialog.getOpenFileNames(
                parent=self,
                caption="Select an RT-DC measurement",
                directory=self.settings.value("paths/add dataset", ""),
                filter="RT-DC Files (*.rtdc)")
        else:
            fnames = paths

        if fnames:
            self.toolButton_new_plot.setEnabled(True)
            self.block_matrix.toolButton_new_plot.setEnabled(True)

        slot_ids = []
        # Create Dataslot instance and update block matrix
        for fn in fnames:
            if is_dcor:
                path = fn
            else:
                path = pathlib.Path(fn)
                self.settings.setValue("paths/add dataset", str(path.parent))
            # add a filter if we don't have one already
            if self.pipeline.num_filters == 0:
                self.add_filter()
            slot_id = self.pipeline.add_slot(path=path)
            self.block_matrix.add_dataset(slot_id=slot_id)
            slot_ids.append(slot_id)

        # Update box filter limits
        self.widget_ana_view.widget_filter.update_box_ranges()
        # Update dataslot in analysis view
        self.widget_ana_view.widget_slot.update_content()
        # redraw
        self.block_matrix.update()
        return slot_ids

    def add_filter(self):
        """Add a filter using tool buttons"""
        filt_id = self.pipeline.add_filter()
        self.block_matrix.add_filter(identifier=filt_id)
        self.widget_ana_view.widget_filter.update_content()
        # redraw
        self.block_matrix.update()
        return filt_id

    def add_plot(self):
        plot_id = self.pipeline.add_plot()
        self.block_matrix.add_plot(identifier=plot_id)
        self.add_plot_window(plot_id)
        # update UI contents
        self.widget_ana_view.widget_plot.update_content()
        # redraw
        self.block_matrix.update()
        return plot_id

    def add_plot_window(self, plot_id):
        """Create a plot window if necessary and show it"""
        if plot_id in self.subwindows_plots:
            sub = self.subwindows_plots[plot_id]
        else:
            # create subwindow
            sub = widgets.MDISubWindowWOButtons(self)
            pw = pipeline_plot.PipelinePlot(parent=sub,
                                            pipeline=self.pipeline,
                                            plot_id=plot_id)
            self.plots_changed.connect(pw.update_content)
            pw.update_content()
            sub.setWidget(pw)
            self.mdiArea.addSubWindow(sub)
            self.subwindows_plots[plot_id] = sub
            sub.setFixedSize(sub.sizeHint())
        sub.show()

    def dragEnterEvent(self, e):
        """Whether files are accepted"""
        if e.mimeData().hasUrls():
            e.accept()
        else:
            e.ignore()

    def dropEvent(self, e):
        """Add dropped files to view"""
        urls = e.mimeData().urls()
        pathlist = []
        for ff in urls:
            pp = pathlib.Path(ff.toLocalFile())
            if pp.is_dir():
                pathlist += list(pp.rglob("*.rtdc"))
            elif pp.suffix == ".rtdc":
                pathlist.append(pp)
        if pathlist:
            self.add_dataslot(paths=sorted(pathlist))

    def init_analysis_view(self):
        sub = widgets.MDISubWindowWOButtons(self)
        self.widget_ana_view = analysis.AnalysisView()
        self.subwindows["analysis_view"] = sub
        sub.setWidget(self.widget_ana_view)
        sub.hide()
        self.mdiArea.addSubWindow(sub)
        self.toolButton_ana_view.clicked.connect(sub.setVisible)
        # applying a new filter triggers updating QuickView
        self.widget_ana_view.widget_filter.pushButton_apply.clicked.connect(
            self.on_quickview_refresh)

    def init_quick_view(self):
        sub = widgets.MDISubWindowWOButtons(self)
        self.widget_quick_view = quick_view.QuickView()
        sub.setWidget(self.widget_quick_view)
        self.toolButton_quick_view.clicked.connect(self.on_quickview)
        self.subwindows["quick_view"] = sub
        # signals
        self.block_matrix.quickviewed.connect(self.on_quickview_show_dataset)
        sub.hide()
        self.mdiArea.addSubWindow(sub)

    def on_action_about(self):
        gh = "ZELLMECHANIK-DRESDEN/ShapeOut2"
        rtd = "shapeout2.readthedocs.io"
        about_text = "Shape-Out 2 is the successor of Shape-Out, " \
            + "a graphical user interface for the analysis and " \
            + "visualization of RT-DC data sets.<br><br>" \
            + "Author: Paul Müller<br>" \
            + "GitHub: " \
            + "<a href='https://github.com/{gh}'>{gh}</a><br>".format(gh=gh) \
            + "Documentation: " \
            + "<a href='https://{rtd}'>{rtd}</a><br>".format(rtd=rtd)
        QtWidgets.QMessageBox.about(self,
                                    "Shape-Out {}".format(__version__),
                                    about_text)

    def on_action_change_dataset_order(self):
        """Show dialog for changing dataset order"""
        dlg = analysis.DlgSlotReorder(self.pipeline, self)
        dlg.pipeline_changed.connect(self.adopt_pipeline)
        dlg.exec()

    @QtCore.pyqtSlot(bool)
    def on_action_check_update(self, b):
        self.settings.setValue("check for updates", int(b))
        if b and self._update_thread is None:
            self._update_thread = QtCore.QThread()
            self._update_worker = update.UpdateWorker()
            self._update_worker.moveToThread(self._update_thread)
            self._update_worker.finished.connect(self._update_thread.quit)
            self._update_worker.data_ready.connect(
                self.on_action_check_update_finished)
            self._update_thread.start()

            version = __version__
            ghrepo = "ZELLMECHANIK-DRESDEN/ShapeOut2"

            QtCore.QMetaObject.invokeMethod(self._update_worker,
                                            'processUpdate',
                                            QtCore.Qt.QueuedConnection,
                                            QtCore.Q_ARG(str, version),
                                            QtCore.Q_ARG(str, ghrepo),
                                            )

    @QtCore.pyqtSlot(dict)
    def on_action_check_update_finished(self, mdict):
        # cleanup
        self._update_thread.quit()
        self._update_thread.wait()
        self._update_worker = None
        self._update_thread = None
        # display message box
        ver = mdict["version"]
        web = mdict["releases url"]
        dlb = mdict["binary url"]
        msg = QtWidgets.QMessageBox()
        msg.setWindowTitle("Shape-Out {} available!".format(ver))
        msg.setTextFormat(QtCore.Qt.RichText)
        text = "You can install Shape-Out {} ".format(ver)
        if dlb is not None:
            text += 'from a <a href="{}">direct download</a>. '.format(dlb)
        else:
            text += 'by running `pip install --upgrade shapeout2`. '
        text += 'Visit the <a href="{}">official release page</a>!'.format(web)
        msg.setText(text)
        msg.exec_()

    def on_action_compute_statistics(self):
        dlg = compute.ComputeStatistics(self, pipeline=self.pipeline)
        dlg.exec()

    def on_action_compute_emodulus(self):
        dlg = bulk.BulkActionEmodulus(self, pipeline=self.pipeline)
        dlg.pipeline_changed.connect(self.adopt_pipeline)
        dlg.exec()

    @QtCore.pyqtSlot()
    def on_action_clear(self, assume_yes=False):
        if assume_yes:
            yes = True
        else:
            buttonReply = QtWidgets.QMessageBox.question(
                self, 'Clear Session', "All progress will be lost. Continue?",
                QtWidgets.QMessageBox.Yes | QtWidgets.QMessageBox.No,
                QtWidgets.QMessageBox.No)
            yes = buttonReply == QtWidgets.QMessageBox.Yes
        if yes:
            session.clear_session(self.pipeline)
            self.reload_pipeline()
        return yes

    def on_action_dcor(self):
        """Show the DCOR import dialog"""
        dlg = dcor.DCORLoader(self)
        dlg.exec()

    def on_action_docs(self):
        webbrowser.open("https://shapeout2.readthedocs.io")

    def on_action_export_data(self):
        dlg = export.ExportData(self, pipeline=self.pipeline)
        if dlg.path is not None:  # user pressed cancel
            dlg.exec()

    def on_action_export_filter(self):
        dlg = export.ExportFilter(self, pipeline=self.pipeline)
        dlg.exec()

    def on_action_export_plot(self):
        dlg = export.ExportPlot(self, pipeline=self.pipeline)
        dlg.exec()

    def on_action_import_filter(self):
        path, _ = QtWidgets.QFileDialog.getOpenFileName(
            self, 'Select Filter', '', 'Filters formats (*.poly *.sof)')
        if path:
            session.import_filters(path, self.pipeline)
            # update UI
            self.reload_pipeline()

    @QtCore.pyqtSlot()
    def on_action_open(self, path=None):
        if self.pipeline.slots or self.pipeline.filters:
            if not self.on_action_clear():
                return
        if path is None:
            path, _ = QtWidgets.QFileDialog.getOpenFileName(
                self, 'Open session', '', 'Shape-Out 2 session (*.so2)',
                'Shape-Out 2 session (*.so2)',
                QtWidgets.QFileDialog.DontUseNativeDialog)
        if path:
            search_paths = []
            while True:
                try:
                    with widgets.ShowWaitCursor():
                        session.open_session(path, self.pipeline, search_paths)
                except session.DataFileNotFoundError as e:
                    missds = "\r".join([str(pp) for pp in e.missing_paths])
                    msg = QtWidgets.QMessageBox()
                    msg.setIcon(QtWidgets.QMessageBox.Warning)
                    msg.setText("Some datasets were not found! "
                                + "Please specify a search location.")
                    msg.setWindowTitle(
                        "Missing {} dataset(s)".format(len(e.missing_paths)))
                    msg.setDetailedText("Missing files: \n\n" + missds)
                    msg.exec_()
                    spath = QtWidgets.QFileDialog.getExistingDirectory(
                        self, 'Data search path')
                    if spath:
                        search_paths.append(spath)
                    else:
                        break
                else:
                    break
            self.reload_pipeline()

    def on_action_preferences(self):
        """Show the DCOR import dialog"""
        dlg = preferences.Preferences(self)
        dlg.exec()

    def on_action_quit(self):
        """Determine what happens when the user wants to quit"""
        if self.pipeline.slots or self.pipeline.filters:
            if not self.on_action_clear():
                return
        QtCore.QCoreApplication.quit()

    def on_action_save(self):
        path, _ = QtWidgets.QFileDialog.getSaveFileName(
            self, 'Save session', '', 'Shape-Out 2 session (*.so2)')
        if path:
            if not path.endswith(".so2"):
                path += ".so2"
            session.save_session(path, self.pipeline)

    def on_action_software(self):
        libs = [dclab,
                h5py,
                numpy,
                pg,
                scipy,
                ]
        sw_text = "Shape-Out {}\n\n".format(__version__)
        sw_text += "Python {}\n\n".format(sys.version)
        sw_text += "Modules:\n"
        for lib in libs:
            sw_text += "- {} {}\n".format(lib.__name__, lib.__version__)
        sw_text += "- PyQt5 {}\n".format(QtCore.QT_VERSION_STR)
        sw_text += "\n Breeze icon theme by the KDE Community (LGPL)."
        if hasattr(sys, 'frozen'):
            sw_text += "\nThis executable has been created using PyInstaller."
        QtWidgets.QMessageBox.information(self,
                                          "Software",
                                          sw_text)

    @widgets.show_wait_cursor
    @QtCore.pyqtSlot()
    def on_data_matrix(self):
        """Show/hide data matrix (User clicked Data Matrix button)"""
        if self.toolButton_dm.isChecked():
            self.splitter.setSizes([200, 1000])
        else:
            self.splitter.setSizes([0, 1])
        # redraw
        self.splitter.update()
        self.mdiArea.update()
        self.block_matrix.update()

    def on_new_polygon_filter(self):
        if not self.pipeline.slots:
            msg = QtWidgets.QMessageBox()
            msg.setIcon(QtWidgets.QMessageBox.Critical)
            msg.setText("A dataset is required for creating a polygon filter!")
            msg.setWindowTitle("No dataset loaded")
            msg.exec_()
        else:
            slot_index, _ = self.block_matrix.get_quickview_indices()
            if slot_index is None:
                # show the first dataset
                self.on_quickview_show_dataset(0, 0)
            self.on_quickview(view=True)
            self.widget_quick_view.on_poly_create()
            # adjusts QuickView size correctly
            self.widget_quick_view.on_tool()

    @widgets.show_wait_cursor
    @QtCore.pyqtSlot(str)
    def on_modify_filter(self, filt_id):
        self.widget_ana_view.tabWidget.setCurrentIndex(2)
        self.widget_ana_view.widget_filter.show_filter(filt_id)
        # finally, check the button
        self.toolButton_ana_view.setChecked(True)
        self.subwindows["analysis_view"].setVisible(True)
        # redraw
        self.mdiArea.update()
        self.subwindows["analysis_view"].update()

    @widgets.show_wait_cursor
    @QtCore.pyqtSlot(str)
    def on_modify_plot(self, plot_id):
        self.widget_ana_view.tabWidget.setCurrentIndex(3)
        self.widget_ana_view.widget_plot.show_plot(plot_id)
        # finally, check the button
        self.toolButton_ana_view.setChecked(True)
        self.subwindows["analysis_view"].setVisible(True)
        # redraw
        self.mdiArea.update()
        self.subwindows["analysis_view"].update()

    @widgets.show_wait_cursor
    @QtCore.pyqtSlot(str)
    def on_modify_slot(self, slot_id):
        self.widget_ana_view.tabWidget.setCurrentIndex(1)
        self.widget_ana_view.widget_slot.show_slot(slot_id)
        # finally, check the button
        self.toolButton_ana_view.setChecked(True)
        self.subwindows["analysis_view"].setVisible(True)
        # redraw
        self.mdiArea.update()
        self.subwindows["analysis_view"].update()

    @widgets.show_wait_cursor
    @QtCore.pyqtSlot(bool)
    def on_quickview(self, view=True):
        """Show/Hide QuickView (User clicked the QuickView button)"""
        self.subwindows["quick_view"].setVisible(view)
        self.block_matrix.enable_quickview(view)
        # redraw
        self.mdiArea.update()
        if view:
            self.subwindows["quick_view"].update()

    @widgets.show_wait_cursor
    @QtCore.pyqtSlot()
    def on_quickview_refresh(self):
        """Refresh quickview with the currently shown dataset"""
        slot_index, filt_index = self.block_matrix.get_quickview_indices()
        if slot_index is not None:
            self.on_quickview_show_dataset(slot_index, filt_index,
                                           update_ana_filter=False)

    @widgets.show_wait_cursor
    @QtCore.pyqtSlot(int, int)
    def on_quickview_show_dataset(self, slot_index, filt_index,
                                  update_ana_filter=True):
        """Update QuickView dataset (User selected new dataset)

        Parameters
        ----------
        slot_index: int
            Index of the slot in `self.pipeline`
        filt_index: int
            Index of the filter in `self.pipeline`
        update_ana_filter: bool
            Whether to update the filter panel in the analysis view.
            If True, matches the filter displayed in the panel to
            the one where QuickView is currently set active. If
            False, nothing is changed.
        """
        ds = self.pipeline.get_dataset(slot_index=slot_index,
                                       filt_index=filt_index,
                                       apply_filter=True)
        slot = self.pipeline.slots[slot_index]
        # update quick view subwindow
        self.widget_quick_view.show_rtdc(ds, slot)
        # show quick view subwindow
        if not self.subwindows["quick_view"].isVisible():
            self.toolButton_quick_view.toggle()
            self.subwindows["quick_view"].setVisible(True)
        if update_ana_filter:
            # update FilterPanel
            filt_id = self.pipeline.filters[filt_index].identifier
            self.widget_ana_view.widget_filter.show_filter(filt_id=filt_id)

    @QtCore.pyqtSlot()
    def on_splitter(self):
        if self.splitter.sizes()[0] == 0:
            self.toolButton_dm.setChecked(False)
        else:
            self.toolButton_dm.setChecked(True)

    def reload_pipeline(self):
        """Convenience function for reloading the current pipeline"""
        self.adopt_pipeline(self.pipeline.__getstate__())


def excepthook(etype, value, trace):
    """
    Handler for all unhandled exceptions.

    :param `etype`: the exception type (`SyntaxError`,
        `ZeroDivisionError`, etc...);
    :type `etype`: `Exception`
    :param string `value`: the exception error message;
    :param string `trace`: the traceback header, if any (otherwise, it
        prints the standard Python header: ``Traceback (most recent
        call last)``.
    """
    vinfo = "Unhandled exception in Shape-Out version {}:\n".format(
        __version__)
    tmp = traceback.format_exception(etype, value, trace)
    exception = "".join([vinfo]+tmp)

    errorbox = QtWidgets.QMessageBox()
    errorbox.setIcon(QtWidgets.QMessageBox.Critical)
    errorbox.addButton(QtWidgets.QPushButton('Close'),
                       QtWidgets.QMessageBox.YesRole)
    errorbox.addButton(QtWidgets.QPushButton(
        'Copy text && Close'), QtWidgets.QMessageBox.NoRole)
    errorbox.setText(exception)
    ret = errorbox.exec_()
    if ret == 1:
        cb = QtWidgets.QApplication.clipboard()
        cb.clear(mode=cb.Clipboard)
        cb.setText(exception)


# Make Ctr+C close the app
signal.signal(signal.SIGINT, signal.SIG_DFL)
# Display exception hook in separate dialog instead of crashing
sys.excepthook = excepthook



import copy

from PyQt5 import QtCore, QtWidgets

from ... import pipeline

from .pm_element import MatrixElement
from .pm_plot import MatrixPlot


class PlotMatrix(QtWidgets.QWidget):
    plot_modify_clicked = QtCore.pyqtSignal(str)
    matrix_changed = QtCore.pyqtSignal()

    def __init__(self, parent=None):
        super(PlotMatrix, self).__init__(parent)

        self.glo = None
        self._reset_layout(init=True)

        # used for toggling between all active, all inactive and semi state
        self.semi_states_plot = {}

    def __getstate__(self):
        """State of the current plot matrix"""
        # plots
        plots = []
        for pw in self.plot_widgets:
            pw_state = pw.__getstate__()
            plot = pipeline.Plot._instances[pw_state["identifier"]]
            plots.append(plot.__getstate__())
        # elements
        mestates = {}
        dm = self.data_matrix
        for dw in dm.dataset_widgets:
            idict = {}
            for pw in self.plot_widgets:
                me = self.get_matrix_element(dw.identifier, pw.identifier)
                idict[pw.identifier] = me.__getstate__()["active"]
            mestates[dw.identifier] = idict
        state = {"elements": mestates,
                 "plots": plots}
        return state

    def __setstate__(self, state):
        self.setUpdatesEnabled(False)
        self.blockSignals(True)
        self.clear()
        # plot states
        for jj in range(len(state["plots"])):
            plot_id = state["plots"][jj]["identifier"]
            pw_state = {"identifier": plot_id,
                        "name": state["plots"][jj]["layout"]["name"],
                        }
            self.add_plot(state=pw_state)
        # make sure elements exist
        self.fill_elements()
        # element states
        for slot_id in state["elements"]:
            ds_d = state["elements"][slot_id]
            for plot_id in ds_d:
                el = self.get_matrix_element(slot_id, plot_id)
                el_state = el.__getstate__()
                el_state["active"] = ds_d[plot_id]
                el.__setstate__(el_state)
        self.blockSignals(False)
        self.setUpdatesEnabled(True)

    def _reset_layout(self, init=False):
        if self.glo is not None:
            # send old layout to Nirvana eventually
            layout = self.glo
            self.glo = None
            self.old_layout = QtWidgets.QWidget()
            self.old_layout.setLayout(layout)
            self.old_layout.hide()
            self.old_layout.deleteLater()
        # add new layout
        self.glo = QtWidgets.QGridLayout()
        self.glo.setSpacing(2)
        self.glo.setContentsMargins(0, 0, 0, 0)
        self.setLayout(self.glo)

    @property
    def data_matrix(self):
        for ch in self.parent().children():
            if ch.__class__.__name__ == "DataMatrix":
                break
        else:
            raise KeyError("DataMatrix not found!")
        return ch

    @property
    def element_widget_dict(self):
        els = {}
        for ii, ws in enumerate(self.data_matrix.dataset_widgets):
            elsd = {}
            for jj, wf in enumerate(self.plot_widgets):
                it = self.glo.itemAtPosition(ii+1, jj)
                elsd[wf.identifier] = it.widget()
            els[ws.identifier] = elsd
        return els

    @property
    def element_width(self):
        """Data matrix element width (without 2px spacing)"""
        return self.data_matrix.element_width

    @property
    def num_datasets(self):
        dm = self.data_matrix
        return dm.num_datasets

    @property
    def num_plots(self):
        count = 0
        for jj in range(self.glo.columnCount()):
            if self.glo.itemAtPosition(0, jj) is not None:
                count += 1
        return count

    @property
    def plot_widgets(self):
        plots = []
        for jj in range(self.glo.columnCount()):
            item = self.glo.itemAtPosition(0, jj)
            if item is not None:
                ps = item.widget()
                plots.append(ps)
        return plots

    def add_plot(self, identifier=None, state=None):
        self.setUpdatesEnabled(False)
        mp = MatrixPlot(identifier=identifier, state=state)
        mp.option_action.connect(self.on_option_plot)
        mp.active_toggled.connect(self.toggle_plot_active)
        mp.modify_clicked.connect(self.plot_modify_clicked.emit)
        self.glo.addWidget(mp, 0, self.num_plots)
        self.fill_elements()
        self.adjust_size()
        self.setUpdatesEnabled(True)
        self.publish_matrix()
        return mp

    def adjust_size(self):
        ncols = self.num_plots
        nrows = self.data_matrix.num_datasets
        if ncols and nrows:
            width1 = self.glo.itemAt(0).widget().width()
            width = (width1 + 2)*ncols - 2
            height = self.data_matrix.sizeHint().height()
            QtWidgets.QApplication.processEvents()
            self.setMinimumSize(width, height)
            self.setFixedSize(width, height)
            QtWidgets.QApplication.processEvents()
            self.setMinimumSize(width, height)
            self.setFixedSize(width, height)

    @QtCore.pyqtSlot()
    def changed_element(self):
        self.publish_matrix()

    def clear(self):
        """Reset layout"""
        self._reset_layout()
        self.semi_states_plot = {}

    def fill_elements(self):
        # add widgets
        for ii in range(self.num_datasets):
            for jj in range(self.num_plots):
                if self.glo.itemAtPosition(ii+1, jj) is None:
                    me = MatrixElement()
                    me.element_changed.connect(self.changed_element)
                    self.glo.addWidget(me, ii+1, jj)
        # make sure enabled/disabled is honored
        dstate = self.data_matrix.__getstate__()
        pstate = self.__getstate__()
        for slot_state in dstate["slots"]:
            slot_id = slot_state["identifier"]
            if slot_id not in dstate["slots used"]:
                for plot_state in pstate["plots"]:
                    plot_id = plot_state["identifier"]
                    me = self.get_matrix_element(slot_id, plot_id)
                    mstate = me.__getstate__()
                    mstate["enabled"] = False
                    me.__setstate__(mstate)

    def get_matrix_element(self, dataset_id, plot_id):
        """Return matrix element matching dataset and plot identifiers"""
        ncols = self.glo.columnCount()
        nrows = self.glo.rowCount()
        for ii in range(1, nrows):
            ds = self.data_matrix.glo.itemAtPosition(ii, 0).widget()
            if ds.identifier == dataset_id:
                for jj in range(ncols):
                    f = self.glo.itemAtPosition(0, jj).widget()
                    if f.identifier == plot_id:
                        break
                else:
                    raise KeyError("Plot '{}' not found!".format(plot_id))
                break
        else:
            raise KeyError("Dataset '{}' not found!".format(dataset_id))
        return self.glo.itemAtPosition(ii, jj).widget()

    def get_plot_index(self, plot_id):
        for ii, pw in enumerate(self.plot_widgets):
            if pw.identifier == plot_id:
                break
        else:
            raise KeyError("Dataset '{}' not found!".format(plot_id))
        return ii

    def get_plot_widget_state(self, plot_id, ret_index=False):
        ii = self.get_plot_index(plot_id)
        pw = self.plot_widgets[ii]
        if ret_index:
            return pw.__getstate__(), ii
        else:
            return pw.__getstate__()

    @QtCore.pyqtSlot(str)
    def on_option_plot(self, option):
        """Plot option logic (remove, duplicate)"""
        pw_state = self.sender().__getstate__()
        plot_id = pw_state["identifier"]
        plot_index = self.get_plot_index(plot_id)
        state = self.__getstate__()
        if option == "remove":
            state["plots"].pop(plot_index)
            for ds_key in state["elements"]:
                state["elements"][ds_key].pop(plot_id)
        else:  # duplicate
            plot = pipeline.Plot()
            new_state = copy.deepcopy(state["plots"][plot_index])
            new_state["identifier"] = plot.identifier
            new_state["layout"]["name"] = plot.name
            state["plots"].insert(plot_index+1, new_state)
            plot.__setstate__(new_state)
        self.__setstate__(state)
        self.publish_matrix()

    def publish_matrix(self):
        """Publish state via self.matrix_changed signal for Pipeline"""
        if not self.signalsBlocked():
            self.matrix_changed.emit()

    @QtCore.pyqtSlot(bool)
    def toggle_dataset_enable(self, enabled):
        sender = self.sender()
        sid = sender.identifier
        state = self.__getstate__()
        for p_key in state["elements"][sid]:
            # update element widget
            me = self.get_matrix_element(sid, p_key)
            mstate = me.__getstate__()
            mstate["enabled"] = enabled
            me.__setstate__(mstate)
        self.publish_matrix()

    @QtCore.pyqtSlot()
    def toggle_plot_active(self):
        """Switch between all active, all inactive, previous state

        Modifies the matrix elements for a plot/column,
        which is defined by the signal sender :class:`MatrixPlot`.
        Cyclic toggling order: semi -> all -> none
        """
        sender = self.sender()
        plot_id = sender.identifier

        states = self.__getstate__()["elements"]
        state = {}
        for slot_id in states:
            state[slot_id] = states[slot_id][plot_id]

        num_actives = sum(list(state.values()))

        # update state according to the scheme in the docstring
        if num_actives == 0:
            if plot_id in self.semi_states_plot:
                # use semi state
                oldstate = self.semi_states_plot[plot_id]
                for slot_id in oldstate:
                    if slot_id in state:
                        state[slot_id] = oldstate[slot_id]
            else:
                # toggle all to active
                for slot_id in state:
                    state[slot_id] = True
        elif num_actives == len(state):
            # toggle all to inactive
            for slot_id in state:
                state[slot_id] = False
        else:
            # save semi state
            self.semi_states_plot[plot_id] = copy.deepcopy(state)
            # toggle all to active
            for slot_id in state:
                state[slot_id] = True

        for slot_id in state:
            me = self.get_matrix_element(slot_id, plot_id)
            me.set_active(state[slot_id])
        self.publish_matrix()

    def update_content(self):
        ncols = self.glo.columnCount()
        nrows = self.glo.rowCount()
        for ii in range(nrows):
            for jj in range(ncols):
                item = self.glo.itemAtPosition(ii, jj)
                if item is not None:
                    item.widget().update_content()



import pkg_resources

import dclab
import numpy as np
from PyQt5 import uic, QtCore, QtWidgets

from ... import meta_tool


class MetaPanel(QtWidgets.QWidget):
    def __init__(self, *args, **kwargs):
        QtWidgets.QWidget.__init__(self)
        path_ui = pkg_resources.resource_filename(
            "shapeout2.gui.analysis", "ana_meta.ui")
        uic.loadUi(path_ui, self)

        self.comboBox_slots.currentIndexChanged.connect(self.update_content)
        self.pipeline_state = None
        self.update_content()

    @property
    def current_slot_state(self):
        if self.pipeline_state is not None:
            slot_index = self.comboBox_slots.currentIndex()
            slot_state = self.pipeline_state["slots"][slot_index]
        else:
            slot_state = None
        return slot_state

    @property
    def slot_ids(self):
        """List of slot identifiers"""
        if self.pipeline_state is not None:
            ids = [ss["identifier"] for ss in self.pipeline_state["slots"]]
        else:
            ids = []
        return ids

    @property
    def slot_names(self):
        """List of slot names"""
        if self.pipeline_state is not None:
            nms = [ss["name"] for ss in self.pipeline_state["slots"]]
        else:
            nms = []
        return nms

    def update_info_box(self, group_box, config, section):
        """Populate an individual group box with keyword-value pairs"""
        group_box.layout().setAlignment(QtCore.Qt.AlignTop)
        # cleanup
        for ii in reversed(range(group_box.layout().count())):
            item = group_box.layout().itemAt(ii).widget()
            if item is not None:
                item.deleteLater()
        # populate
        items = sort_config_section_items(section, config[section].items())
        if items:
            for key, value in items:
                k, v, t = format_config_key_value(section, key, value)
                widget = QtWidgets.QWidget()
                widget.setToolTip(t)
                hbox = QtWidgets.QHBoxLayout()
                hbox.setAlignment(QtCore.Qt.AlignLeft)
                hbox.setContentsMargins(0, 0, 0, 0)
                hbox.addWidget(QtWidgets.QLabel(k + ": "))
                hbox.addWidget(QtWidgets.QLabel(v))
                widget.setLayout(hbox)
                group_box.layout().addWidget(widget)
            group_box.show()
        else:
            group_box.hide()

    def update_content(self, event=None, slot_index=None):
        if self.slot_ids:
            self.setEnabled(True)
            # update combobox
            self.comboBox_slots.blockSignals(True)
            if slot_index is None:
                slot_index = self.comboBox_slots.currentIndex()
                if slot_index > len(self.slot_ids) - 1 or slot_index < 0:
                    slot_index = len(self.slot_ids) - 1
            self.comboBox_slots.clear()
            self.comboBox_slots.addItems(self.slot_names)
            self.comboBox_slots.setCurrentIndex(slot_index)
            self.comboBox_slots.blockSignals(False)
            # populate content
            slot_state = self.pipeline_state["slots"][slot_index]
            cfg = meta_tool.get_rtdc_config(slot_state["path"])
            self.update_info_box(self.groupBox_experiment, cfg,
                                 "experiment")
            self.update_info_box(self.groupBox_fluorescence, cfg,
                                 "fluorescence")
            self.update_info_box(self.groupBox_imaging, cfg,
                                 "imaging")
            self.update_info_box(self.groupBox_online_contour, cfg,
                                 "online_contour")
            self.update_info_box(self.groupBox_online_filter, cfg,
                                 "online_filter")
            self.update_info_box(self.groupBox_setup, cfg,
                                 "setup")
        else:
            self.setEnabled(False)


def format_config_key_value(section, key, value):
    mdsec = dclab.dfn.CFG_METADATA[section]
    for dckey, dctype, descr in mdsec:
        if dckey == key:
            break
    else:
        raise KeyError("Unknwon key [{}]: {}".format(section, key))
    tip = ""
    # Value formatting
    if dctype == float:  # pretty-print floats
        if value == 0:
            string = "0.0"
        else:
            # determine number of decimals
            dec = int(np.ceil(np.log10(1/np.abs(value))))
            if dec < 0:
                dec = 0
            string = ("{:." + "{}".format(dec + 2) + "f}").format(value)
    else:
        string = str(value)

    # Special cases
    if section == "experiment":
        if key in ["date", "time"]:
            descr, form = descr.split("(")
            tip = form.strip("()'")
        elif key == "sample":
            descr = "Sample name"
    elif section == "setup":
        if key == "chip region":
            descr = descr.split(" (")[0]
        elif key == "medium" and string == "CellCarrierB":
            string = "CellCarrier B"
        elif key == "module composition":
            descr = "Modules used"
            string = ", ".join(string.split(","))
        elif key == "software version":
            descr = "Software"

    # Units
    if descr.endswith("]"):
        descr, units = descr.rsplit(" [", 1)
        units = units.strip("] ")
        string += " " + units

    return descr, string, tip


def sort_config_section_items(section, items):
    if section == "experiment":
        order = ["sample", "run index", "event count", "date", "time"]
    elif section == "setup":
        order = ["medium", "channel width"]
    else:
        order = None

    if order is None:
        sitems = items
    else:
        sitems = []
        for key in order:
            for item in items:
                if key == item[0]:
                    sitems.append(item)
                    break
        # append those not in `order`
        for item in items:
            if item not in sitems:
                sitems.append(item)
    return sitems



"""Collect relevant icons to icon theme subdirectories

The icons used rely on the KDE breeze theme.

    apt install breeze-icon-theme

This script must be run on a linux machine. Please make sure
that `icon_root` is correct.
"""
import pathlib
import shutil

# The key identifies the theme; each list contains icon names
icons = {
    "breeze": [
        "application-exit",
        "code-context",
        "dialog-cancel",
        "dialog-close",
        "dialog-error",
        "dialog-information",
        "dialog-messages",
        "dialog-ok",
        "dialog-ok-apply",
        "dialog-question",
        "dialog-warning",
        "documentinfo",
        "document-open",
        "document-open-folder",
        "document-save",
        "draw-watercolor",
        "edit-clear",
        "edit-paste",
        "folder",
        "folder-cloud",
        "globe",
        "gtk-preferences",
        "list-add",
        "messagebox_warning",
        "object-columns",
        "object-order-lower",
        "object-rows",
        "office-chart-ring",
        "office-chart-scatter",
        "remove",
        "search",
        "show-grid",
        "special_paste",
        "tools-wizard",
        "view-filter",
        "view-statistics",
        "visibility",
        "path-mode-polyline",
        "preferences-activities",
    ],
}

# theme index file
index = """[Icon Theme]
Name=ShapeOutMix
Comment=Mix of themes for Shape-Out 2

Directories={directories}
"""

# theme file folder item
index_item = """
[{directory}]
Size={res}
Type=Fixed
"""

icon_root = pathlib.Path("/usr/share/icons")


def find_icons(name, theme):
    cands = sorted((icon_root / theme).rglob("{}.svg".format(name)))
    cands += sorted((icon_root / theme).rglob("{}.png".format(name)))
    return cands


if __name__ == "__main__":
    directories = []
    here = pathlib.Path(__file__).parent
    for theme in icons:
        for name in icons[theme]:
            ipaths = find_icons(name, theme)
            if not ipaths:
                print("Could not find {} {}".format(theme, name))
                continue
            for ipath in ipaths:
                relp = ipath.parent.relative_to(icon_root)
                dest = here / relp
                directories.append(str(relp))
                dest.mkdir(exist_ok=True, parents=True)
                shutil.copy(ipath, dest)

    with (here / "index.theme").open("w") as fd:
        directories = sorted(set(directories))
        fd.write(index.format(directories=",".join(
            ["shapeout2"] + directories)))
        # Shape-Out icons
        fd.write(index_item.format(directory="shapeout2", res="16"))
        # theme icons
        for dd in directories:
            for res in ["16", "22", "24", "32", "64", "128"]:
                if res in str(dd):
                    break
            else:
                raise ValueError("No resolution for {}!".format(dd))
            fd.write(index_item.format(directory=dd, res=res))



import itertools
import pathlib
import shutil
import tempfile

import h5py
import numpy as np
from shapeout2 import pipeline, session


def equal_state(a, b):
    """Compares dictionaries recursively

    - nan-values are allowed
    """
    if isinstance(a, dict):
        equal = True
        for key in a:
            equal &= equal_state(a[key], b[key])
    elif isinstance(a, list):
        equal = True
        for ai, bi in zip(a, b):
            equal &= equal_state(ai, bi)
    elif isinstance(a, float):
        if np.isnan(a):
            equal = np.isnan(b)
        else:
            equal = not np.isnan(a)
    else:
        equal = a == b
    return equal


def make_pipeline(nslots=2, nfilters=3, nplots=1, paths=None):
    """Create a sample pipeline

    - Each filter has a different box filter
    - Every other filter/plot will be set active
    """
    if paths is None:
        here = pathlib.Path(__file__).parent
        paths = [here / "data" / "calibration_beads_47.rtdc"]

    # circular iterator over paths
    pathcycle = itertools.cycle(paths)

    # initiate the pipeline
    pl = pipeline.Pipeline()

    for _ in range(nslots):
        pl.add_slot(path=next(pathcycle))

    feats = pl.get_features(scalar=True)

    for jj in range(nfilters):
        filt_id = pl.add_filter()
        filt = pl.get_filter(filt_id)

        feat = feats[jj]
        fmin, fmax = pl.get_min_max(feat)
        # modify the filter
        filt.boxdict[filt_id] = {"start": fmin,
                                 "end": (fmin + fmax)/2,
                                 "active": True}

    # set every other filter active
    active = True
    for slot_id in pl.slot_ids:
        for filt_id in pl.filter_ids:
            pl.set_element_active(slot_id, filt_id, active)
            active = not active

    # add a plot
    for _ in range(nplots):
        pl.add_plot()

    # set every other plot active
    active = True
    for slot_id in pl.slot_ids:
        for plot_id in pl.plot_ids:
            pl.set_element_active(slot_id, plot_id, active)
            active = not active

    return pl


def test_2_1_1_new_key_emodulus_enabled():
    """In Shape-Out 2.1.1 we introduces the "emodulus enabled" key

    If it is disabled (reservoir measurements), then the emodulus
    analysis options are not shown in the Slot options. See also
    changes made in dclab 0.22.4 (test for reservoir existence).
    """
    spath = pathlib.Path(__file__).parent / "data" / "version_2_1_0_basic.so2"
    pl = session.open_session(spath)
    sc = pl.slots[0].config
    assert "emodulus" in sc
    assert "emodulus enabled" in sc["emodulus"]
    assert sc["emodulus"]["emodulus enabled"]


def test_file_hash():
    tempdir = pathlib.Path(tempfile.mkdtemp(prefix="test_shapeout2_session_"))
    # custom path to measurement
    p0 = pathlib.Path(__file__).parent / "data" / "calibration_beads_47.rtdc"
    pp = tempdir / "calibration_beads_47.rtdc"
    shutil.copy(p0, pp)
    hash1 = session.hash_file_partially(pp)
    session.hash_file_partially.cache_clear()  # force recomputation of hashes
    # modify the file
    with h5py.File(pp, mode="a") as h5:
        h5.attrs["setup:medium"] = "unknown"
    hash2 = session.hash_file_partially(pp)
    assert hash1 != hash2
    # cleanup
    shutil.rmtree(tempdir, ignore_errors=True)


def test_missing_path_in_session():
    tempdir = pathlib.Path(tempfile.mkdtemp(prefix="test_shapeout2_session_"))
    spath = tempdir / "session.so2"
    # custom path to measurement
    p0 = pathlib.Path(__file__).parent / "data" / "calibration_beads_47.rtdc"
    pp = tempdir / "calibration_beads_47.rtdc"
    shutil.copy(p0, pp)
    pl = make_pipeline(paths=[pp])
    session.save_session(spath, pl)
    session.clear_session(pl)
    # rename the file
    pc = pp.with_name("calibration_beads_47_moved.rtdc")
    pp.rename(pc)
    # load bad session
    try:
        session.open_session(spath)
    except session.DataFileNotFoundError as e:
        assert pp in e.missing_paths
    else:
        assert False, "should have raised an error!"
    # try again with proper search path
    pl3 = session.open_session(spath, search_paths=[pc])
    session.clear_session(pl3)
    # try again with a directory as search path
    other = tempdir / "other"
    other.mkdir()
    pc.rename(other / pp.name)  # must have same name as `pp`
    pl4 = session.open_session(spath, search_paths=[other])
    session.clear_session(pl4)
    # cleanup
    shutil.rmtree(tempdir, ignore_errors=True)


def test_relative_paths():
    tempdir = pathlib.Path(tempfile.mkdtemp(prefix="test_shapeout2_session_"))
    # custom path for data
    p0 = pathlib.Path(__file__).parent / "data" / "calibration_beads_47.rtdc"
    datadir = tempdir / "data"
    datadir.mkdir()
    pp = datadir / "calibration_beads_47.rtdc"
    shutil.copy(p0, pp)
    # custom path for session
    sessiondir = tempdir / "session"
    sessiondir.mkdir()
    spath = sessiondir / "session.so2"
    # pipeline
    pl = make_pipeline(paths=[pp])
    session.save_session(spath, pl)
    session.clear_session(pl)
    # new session directory
    new_sessiondir = tempdir / "new" / "path" / "abracadabra"
    new_sessiondir.mkdir(parents=True)
    new_spath = new_sessiondir / spath.name
    spath.rename(new_spath)
    # new path directory (same relative path)
    new_datadir = tempdir / "new" / "path" / "data"
    new_datadir.mkdir(parents=True)
    new_pp = new_datadir / pp.name
    pp.rename(new_pp)
    # and load it (without search_paths as arguments)
    session.open_session(new_spath)
    # cleanup
    shutil.rmtree(tempdir, ignore_errors=True)


def test_simple_save_open_session():
    pl = make_pipeline()
    old_state = pl.__getstate__()

    tempdir = pathlib.Path(tempfile.mkdtemp(prefix="test_shapeout2_session_"))
    spath = tempdir / "session.so2"

    session.save_session(spath, pl)

    # currently, there may only be one pipeline
    session.open_session(spath, pl)
    new_state = pl.__getstate__()

    assert equal_state(old_state, new_state)

    # test opposite
    old_state["slots"][0]["emodulus"]["emodulus temperature"] = 10
    assert not equal_state(old_state, new_state)
    # cleanup
    shutil.rmtree(tempdir, ignore_errors=True)


def test_wrong_hash():
    tempdir = pathlib.Path(tempfile.mkdtemp(prefix="test_shapeout2_session_"))
    spath = tempdir / "session.so2"
    # custom path to measurement
    p0 = pathlib.Path(__file__).parent / "data" / "calibration_beads_47.rtdc"
    pp = pathlib.Path(tempdir) / "calibration_beads_47.rtdc"
    shutil.copy(p0, pp)
    pl = make_pipeline(paths=[pp])
    session.save_session(spath, pl)
    session.clear_session(pl)  # clear session before opening h5 file rw
    session.hash_file_partially.cache_clear()  # force recomputation of hashes
    # modify the file
    with h5py.File(pp, mode="a") as h5:
        h5.attrs["setup:medium"] = "unknown"
    # opening modified file should just work if the path matches
    pl2 = session.open_session(spath)
    session.clear_session(pl2)
    # but when the directory changes, the hash is checked
    other = tempdir / "other"
    other.mkdir()
    pp.rename(other / pp.name)
    try:
        session.open_session(spath, search_paths=[other])
    except session.DataFileNotFoundError as e:
        assert pp in e.missing_paths
    else:
        assert False, "should have raised an error!"
    # cleanup
    shutil.rmtree(tempdir, ignore_errors=True)


if __name__ == "__main__":
    # Run all tests
    loc = locals()
    for key in list(loc.keys()):
        if key.startswith("test_") and hasattr(loc[key], "__call__"):
            loc[key]()



import shutil
import tempfile
import time

from PyQt5 import QtCore


TMPDIR = tempfile.mkdtemp(prefix=time.strftime(
    "shapeout2_test_%H.%M_"))


def pytest_configure(config):
    """This is ran before all tests"""
    # disable update checking
    QtCore.QCoreApplication.setOrganizationName("Zellmechanik-Dresden")
    QtCore.QCoreApplication.setOrganizationDomain("zellmechanik.com")
    QtCore.QCoreApplication.setApplicationName("shapeout2")
    QtCore.QSettings.setDefaultFormat(QtCore.QSettings.IniFormat)
    settings = QtCore.QSettings()
    settings.setIniCodec("utf-8")
    settings.setValue("check for updates", 0)
    settings.setValue("advanced/check pyqtgraph version", 0)
    settings.sync()
    # set global temp directory
    tempfile.tempdir = TMPDIR


def pytest_unconfigure(config):
    """
    called before test process is exited.
    """
    # clear global temp directory
    shutil.rmtree(TMPDIR, ignore_errors=True)



import dclab


class FilterRay(object):
    def __init__(self, slot):
        """Manages filter-based dataset hierarchies

        Filter rays are used to cache RTDCBase filter-hierarchy
        children.
        """
        #: identifier of the ray (matches the slot)
        self.identifier = slot.identifier
        #: slot defining the ray
        self.slot = slot
        #: list of RTDCBase (hierarchy children)
        self.steps = []
        #: corresponds to hashes of the applied filters
        self.step_hashes = []
        # holds the filters (protected so that users use set_filters)
        self._filters = []
        # used for testing (incremented when the ray is cut)
        self._generation = 0
        # used for checking validity of the ray
        self._slot_hash = "unset"
        self._root_child = None

    def __repr__(self):
        repre = "<Pipeline Filter Ray '{}' at {}>".format(self.identifier,
                                                          hex(id(self)))
        return repre

    def _add_step(self, ds, filt):
        """Add a filter step"""
        self.step_hashes.append(filt.hash)
        filt.update_dataset(ds)
        self.steps.append(ds)
        return self._new_child(ds, filt)

    def _new_child(self, ds, filt=None, apply_filter=False):
        identifier = self.slot.identifier
        if filt is None:
            identifier += "-root"
        else:
            identifier += "-" + filt.identifier + "-child"
        ds = dclab.rtdc_dataset.RTDC_Hierarchy(
            ds, apply_filter=apply_filter, identifier=identifier)
        return ds

    @property
    def filters(self):
        """filters currently used by the ray"""
        return self._filters

    @property
    def root_child(self):
        """This is the first element in self.steps
        (Will return a dataset even if self.steps is empty)
        """
        if self._slot_hash != self.slot.hash:
            # reset everything (e.g. emodulus recipe might have changed)
            self.steps = []
            self.step_hashes = []
            self._root_child = self._new_child(self.slot.get_dataset(),
                                               apply_filter=True)
            self._slot_hash = self.slot.hash
        return self._root_child

    def get_final_child(self, rtdc_ds=None, apply_filter=True):
        """Return the final ray child of `rtdc_ds`

        If `rtdc_ds` is None, then the dataset of the current
        ray (self.slot) is used. If `rtdc_ds` is given, then
        no ray caching is performed and the present ray is not
        modified.

        This is a convenience function used when the filter ray
        must be applied to a different dataset (not the one in
        `self.slot`). This is used in Shape-Out when a filter ray
        is applied to other data on disk e.g. when computing
        statistics. For regular use of the filter ray in a
        pipeline, use :func:`get_dataset`.
        """
        filters = self.filters

        if rtdc_ds is None:
            # normal case
            external = False
            rtdc_ds = self.slot.get_dataset()
            ds = self.root_child
        else:
            # ray is applied to other data
            external = True
            # do not modify rtdc_ds (create a child to work with)
            ds = self._new_child(rtdc_ds, apply_filter=True)

        # Dear future self,
        #
        # don't even think about filter ray branching.
        #
        # Sincerely,
        # past self

        if filters:
            # apply all filters
            for ii, filt in enumerate(filters):
                # remmeber the previous hierarchy parent
                # (ds is always used for the next iteration)
                prev_ds = ds
                if external:
                    # do not touch self.steps or self.step_hashes
                    filt.update_dataset(ds)
                    ds = self._new_child(ds, filt)
                elif len(self.steps) < ii+1:
                    # just create a new step
                    ds = self._add_step(ds, filt)
                elif filt.hash != self.step_hashes[ii]:
                    # the filter ray is changing here;
                    # cut it and add a new step
                    self.steps = self.steps[:ii]
                    self.step_hashes = self.step_hashes[:ii]
                    ds = self._add_step(ds, filt)
                    self._generation += 1  # for testing
                else:
                    # the filters match so far
                    if len(self.steps) > ii + 1:  # next child exists
                        ds = self.steps[ii + 1]
                    else:  # next child does not exist
                        ds = self._new_child(ds, filt)
            # we now have the entire filter pipeline in self.steps
            final_ds = prev_ds
        else:
            final_ds = rtdc_ds
        if apply_filter:
            final_ds.apply_filter()
        return final_ds

    def get_dataset(self, filters=None, apply_filter=True):
        """Return the dataset that corresponds to applying these filters

        Parameters
        ----------
        filters: list of Filter or None
            Filters used for computing the dataset hierarchy. If set
            to None, the current filters in `self.filters` are used.
        """
        if filters is not None:
            # put the filters in place
            self.set_filters(filters)
        # compute the final hierarchy child
        ds = self.get_final_child(apply_filter=apply_filter)
        return ds

    def set_filters(self, filters):
        """Set the filters of the current ray"""
        # only take into account active filters
        self._filters = [f for f in filters if f.filter_used]



import copy
import warnings

import dclab
import numpy as np

from .dataslot import Dataslot
from .filter import Filter
from .filter_ray import FilterRay
from .plot import Plot


class EmptyDatasetWarning(UserWarning):
    pass


class MissingFeatureWarning(UserWarning):
    pass


class Pipeline(object):
    def __init__(self, state=None):
        self.reset()
        #: previous state (see __setstate__)
        self._old_state = {}
        if state is not None:
            self.__setstate__(state)

    def __setstate__(self, state):
        if self._old_state == state:
            # Nothing changed
            return
        self._old_state = state
        self.reset()
        for filt_state in state["filters"]:
            self.add_filter(filt_state)
        for plot_state in state["plots"]:
            self.add_plot(plot_state)
        for slot_state in state["slots"]:
            self.add_slot(slot=slot_state)
        # set element states at the end
        self.element_states = state["elements"]

        # sanity checks
        if set(self.filters_used) != set(state["filters used"]):
            raise ValueError("Bad pipeline state ('filters used' don't match)")
        if set(self.slots_used) != set(state["slots used"]):
            raise ValueError("Bad pipeline state ('slots used' don't match)")

    def __getstate__(self):
        state = {}
        state["elements"] = copy.deepcopy(self.element_states)
        state["filters"] = [filt.__getstate__() for filt in self.filters]
        state["filters used"] = self.filters_used
        state["plots"] = [plot.__getstate__() for plot in self.plots]
        state["slots"] = [slot.__getstate__() for slot in self.slots]
        state["slots used"] = self.slots_used
        return state

    @property
    def filter_ids(self):
        return [filt.identifier for filt in self.filters]

    @property
    def filters_used(self):
        return [filt.identifier for filt in self.filters if filt.filter_used]

    @property
    def plot_ids(self):
        return [plot.identifier for plot in self.plots]

    @property
    def slot_ids(self):
        return [slot.identifier for slot in self.slots]

    @property
    def slots_used(self):
        return [slot.identifier for slot in self.slots if slot.slot_used]

    @property
    def num_filters(self):
        return len(self.filters)

    @property
    def num_plots(self):
        return len(self.plots)

    @property
    def num_slots(self):
        return len(self.slots)

    @property
    def paths(self):
        return [ds.path for ds in self.slots]

    def add_filter(self, filt=None, index=None):
        """Add a filter to the pipeline

        Parameters
        ----------
        filt: shapeout2.pipeline.Filter or dict
            Filter instance or its state from Filter.__getstate__()
        index: int
            Position in the filter list, defaults to `len(self.filters)`

        Returns
        -------
        index: int
            index of the filter in the pipeline;
            indexing starts at "0".
        """
        if index is None:
            index = self.num_filters
        if filt is None:
            filt = Filter()
        elif isinstance(filt, Filter):
            pass
        elif isinstance(filt, dict):
            state = filt
            if state["identifier"] in Filter._instances:
                filt = Filter._instances[filt["identifier"]]
            else:
                filt = Filter()
            filt.__setstate__(state)
        self.filters.insert(index, filt)
        filt_id = filt.identifier
        for slot_id in self.slot_ids:
            self.element_states[slot_id][filt_id] = False
        return filt_id

    def add_plot(self, plot=None, index=None):
        """Add a plot to the pipeline

        Parameters
        ----------
        plot: shapeout2.pipeline.Plot or dict
            Plot instance or its state from Plot.__getstate__()
        index: int
            Position in the plot list, defaults to `len(self.plots)`

        Returns
        -------
        index: int
            index of the plot in the pipeline;
            indexing starts at "0".
        """
        if index is None:
            index = self.num_plots
        if plot is None:
            plot = Plot()
        elif isinstance(plot, Plot):
            pass
        elif isinstance(plot, dict):
            state = plot
            if state["identifier"] in Plot._instances:
                plot = Plot._instances[plot["identifier"]]
            else:
                plot = Plot()
            plot.__setstate__(state)

        self.plots.insert(index, plot)
        plot_id = plot.identifier
        for slot_id in self.slot_ids:
            self.element_states[slot_id][plot_id] = False

        return plot.identifier

    def add_slot(self, slot=None, path=None, index=None):
        """Add a slot (experiment) to the pipeline

        Parameters
        ----------
        slot: shapeout2.pipeline.Dataslot or dict
            Dataslot representing an experimental dataset or its
            state from Dataslot.__getstate__(); At least `slot`
            or `path` need to be specified
        path: str or pathlib.Path
            Path to a measurement
        index: int
            Position in the slot list, defaults to `len(self.slots)`

        Returns
        -------
        index: int
            index of the slot in the pipeline;
            indexing starts at "0".
        identifier: str
            identifier of the slot
        """
        if index is None:
            index = self.num_slots
        if ((slot is None and path is None)
                or (slot is not None and path is not None)):
            raise ValueError("Please specify either `slot` or `path`.")
        elif path is not None:
            slot = Dataslot(path=path)
        elif isinstance(slot, Dataslot):
            pass
        elif isinstance(slot, dict):
            state = slot
            if state["identifier"] in Dataslot._instances:
                slot = Dataslot._instances[slot["identifier"]]
            else:
                slot = Dataslot()
            slot.__setstate__(state)

        self.slots.insert(index, slot)
        slot_id = slot.identifier
        self.element_states[slot_id] = {}
        for filt_id in self.filter_ids:
            self.element_states[slot_id][filt_id] = False
        for plot_id in self.plot_ids:
            self.element_states[slot_id][plot_id] = False
        return slot.identifier

    def apply_filter_ray(self, rtdc_ds, slot_id):
        """Convenience function for applying filters to other data

        The filters that are currently set for a specific slot with
        `slot_id` are applied to a dataset.

        Parameters
        ----------
        rtdc_ds: dclab.rtdc_dataset.RTDCBase
            Dataset
        slot_id: str
            Identifier of the slot from which the filters are taken
        """
        # make sure the current ray is built correctly
        self.get_dataset(self.slot_ids.index(slot_id), apply_filter=False)
        # get the ray
        ray = self.get_ray(slot_id)
        ds = ray.get_final_child(rtdc_ds)
        return ds

    def get_dataset(self, slot_index, filt_index=-1, apply_filter=True):
        """Return dataset with all filters updated (optionally applied)

        Parameters
        ----------
        slot_index: int
            index of measurement
        filt_index: int or None
            index of filter; if None, then the plain dataset is returned.
            If negative (default), then the last dataset in the pipeline
            is returned (all selected filters applied).
        apply_filter: bool
            whether to call `dataset.apply_filter` in the end;
            if set to `False`, only the filtering configuration
            of the dataset and its hierarchy parents are updated
        ret_color: bool
            also return the color of the dataset as a string
        """
        if not isinstance(slot_index, int):
            raise ValueError(
                "`slot_index` must be an integer, got '{}'".format(slot_index))
        slot = self.slots[slot_index]
        if filt_index is None or (filt_index == -1 and len(self.slots) == 0):
            # return the unfiltered dataset
            ds = slot.get_dataset()
        else:
            if filt_index < 0:
                filt_index = len(self.filters) - 1
            slot_id = slot.identifier
            # the filters used
            filters = []
            for filt_id in self.filter_ids[:filt_index+1]:
                if (self.is_element_active(slot_id, filt_id)
                        and filt_id in self.filters_used):
                    filters.append(self.get_filter(filt_id))
            # filter ray magic
            ray = self.get_ray(slot.identifier)
            ds = ray.get_dataset(filters, apply_filter=apply_filter)
        return ds

    def get_datasets(self, filt_index=-1, apply_filter=True):
        """Return all datasets with filters applied

        The parameters are passed to :func:`Pipeline.get_dataset`.
        """
        kw = {"filt_index": filt_index,
              "apply_filter": apply_filter}
        return [self.get_dataset(ii, **kw) for ii in range(len(self.slots))]

    def get_features(self, scalar=False, label_sort=False, union=False,
                     plot_id=None, ret_labels=False):
        """Return a list of features in the pipeline

        Parameters
        ----------
        scalar: bool
            If True, only return scalar features
        label_sort: bool
            If True, return the features sorted by label
            instead of by feature name
        union: bool
            If True, return the union of features available in all
            slots. If False (default), return only those features
            that are shared by all slots.
        plot_id: None or str
            If set, only datasets that are part of the matching Plot
            instance are used. If None, all datasets of the pipeline
            are used.
        ret_labels: bool
            If True, return the labels as well

        Notes
        -----
        This function returns an empty list if there are no features
        available.
        """
        if union:
            features = set()
        else:
            features = None
        for slot_index in range(self.num_slots):
            slot_id = self.slot_ids[slot_index]
            if (plot_id is None
                or (self.element_states[slot_id][plot_id]
                    and slot_id in self.slots_used)):
                ds = self.get_dataset(slot_index=slot_index, filt_index=None)
                if scalar:
                    ds_features = set(ds.features_scalar)
                else:
                    ds_features = set(ds.features)
                if union:
                    features |= ds_features
                else:
                    if features is None:
                        features = ds_features
                    else:
                        features &= ds_features
        if features is None:
            # This means that the pipeline is empty
            features = []
        labs = [dclab.dfn.get_feature_label(f) for f in features]
        if label_sort:
            lf = sorted(zip(labs, features))
            features = [it[1] for it in lf]
            labs = [it[0] for it in lf]
        else:
            fl = sorted(zip(features, labs))
            features = [it[0] for it in fl]
            labs = [it[1] for it in fl]
        if ret_labels:
            return features, labs
        else:
            return features

    def get_filter(self, filt_id):
        """Return the Filter matching the identifier"""
        if filt_id not in self.filter_ids:
            raise ValueError(
                "Filter '{}' not part of this pipeline!".format(filt_id))
        return self.filters[self.filter_ids.index(filt_id)]

    def get_min_max(self, feat, plot_id=None, margin=0.0):
        """Return minimum and maximum values for a feature

        Parameters
        ----------
        feat: str
            Feature name
        plot_id: str
            Plot identifier
        margin: float
            Fraction by which the minimum and maximum are
            extended. E.g. for plotting with a 5% margin
            use `margin=0.05`.

        Returns
        -------
        [fmin, fmax]: list of float
            Minimum and maximum values of the feature. If the feature
            is empty or only-nan, an :class:`EmptyDatasetWarning` is
            issued and both return values are set to zero.
        """
        if plot_id is not None:
            dslist = self.get_plot_datasets(plot_id)[0]
        else:
            dslist = self.get_datasets(filt_index=None, apply_filter=False)
        fmin = np.inf
        fmax = -np.inf
        for ds in dslist:
            if np.sum(ds.filter.all):
                if feat in ds:
                    fdata = ds[feat][ds.filter.all]
                    invalid = np.logical_or(np.isnan(fdata), np.isinf(fdata))
                    vdata = fdata[~invalid]
                    vmin = np.min(vdata)
                    vmax = np.max(vdata)
                    fmin = np.min([fmin, vmin])
                    fmax = np.max([fmax, vmax])
                else:
                    warnings.warn("Dataset {} does not ".format(ds.identifier)
                                  + "contain the feature '{}'!".format(feat),
                                  MissingFeatureWarning)
            else:
                warnings.warn("Dataset {} does not ".format(ds.identifier)
                              + "contain any events when filtered!",
                              EmptyDatasetWarning)
        if margin:
            diff = fmax - fmin
            fmin -= margin*diff
            fmax += margin*diff

        if np.any(np.isinf([fmin, fmax])):
            # Set values to 0 if no
            fmin = fmax = 0
        return [fmin, fmax]

    def get_plot(self, plot_id):
        if plot_id not in self.plot_ids:
            raise ValueError(
                "Plot '{}' not part of this pipeline!".format(plot_id))
        return self.plots[self.plot_ids.index(plot_id)]

    def get_plot_datasets(self, plot_id, apply_filter=True):
        """Return a list of datasets with slot states that belong to a plot"""
        datasets = []
        states = []
        # keep the same order as in self.slots
        for slot_index in range(len(self.slots)):
            slot = self.slots[slot_index]
            slot_id = slot.identifier
            if (self.element_states[slot_id][plot_id]
                    and slot_id in self.slots_used):
                ds = self.get_dataset(slot_index=slot_index,
                                      apply_filter=apply_filter)
                datasets.append(ds)
                states.append(slot.__getstate__())
        return datasets, states

    def get_plot_col_row_count(self, plot_id, pipeline_state=None):
        """Compute how many rows a plot layout requires

        Parameters
        ----------
        plot_id: str
            identifier of a plot in this pipeline
        pipeline_state: dict
            pipeline state to use; defaults to the current pipeline
            state
        """
        if pipeline_state is None:
            pipeline_state = self.__getstate__()

        # plot state
        for pstate in pipeline_state["plots"]:
            if plot_id == pstate["identifier"]:
                break
        else:
            raise KeyError(
                "Plot '{}' not given in pipeline state!".format(plot_id))

        # number of datasets in that plot
        num_scat = 0
        for slot_id in pipeline_state["elements"]:
            num_scat += pipeline_state["elements"][slot_id][plot_id]

        # additional plots
        div = pstate["layout"]["division"]
        if div == "each":
            num_plots = max(1, num_scat)
        elif div == "merge":
            num_plots = 1
        elif div == "multiscatter+contour":
            num_plots = num_scat + 1
        else:
            raise ValueError("Unrecognized division: '{}'".format(div))

        # column count
        col_count = min(pstate["layout"]["column count"], num_plots)

        # row count
        row_count = int(np.ceil(num_plots/col_count))
        return col_count, row_count

    def get_ray(self, slot_id):
        """Convenience function that creates and returns a filter ray"""
        # cleanup (just in case)
        for key in list(self.rays.keys()):
            if key not in self.slot_ids:
                self.rays.pop(key)
        if slot_id not in self.rays:
            self.rays[slot_id] = FilterRay(self.get_slot(slot_id))
        return self.rays[slot_id]

    def get_slot(self, slot_id):
        """Return the Dataslot matching the RTDCBase identifier"""
        slot_id = slot_id.split("-")[0]  # this is how FilterRay names children
        if slot_id in self.slot_ids:
            slot = self.slots[self.slot_ids.index(slot_id)]
        else:
            raise ValueError("Unknown dataset identifier: "
                             + "`{}`".format(slot_id))
        return slot

    def is_element_active(self, slot_id, filt_plot_id):
        return self.element_states[slot_id][filt_plot_id]

    def remove_filter(self, filt_id):
        """Remove a filter by filter identifier"""
        index = self.filter_ids.index(filt_id)
        self.filters.pop(index)
        for slot_id in self.element_states:
            if filt_id in self.element_states[slot_id]:
                self.element_states[slot_id].pop(filt_id)

    def remove_plot(self, plot_id):
        """Remove a filter by plot identifier"""
        index = self.plot_ids.index(plot_id)
        self.plots.pop(index)
        for slot_id in self.element_states:
            if plot_id in self.element_states[slot_id]:
                self.element_states[slot_id].pop(plot_id)

    def remove_slot(self, slot_id):
        """Remove a slot by slot identifier"""
        index = self.slot_ids.index(slot_id)
        self.slots.pop(index)
        if slot_id in self.element_states:
            self.element_states.pop(slot_id)

    def reorder_slots(self, indices):
        """Change the order of data slots

        Parameters
        ----------
        indices: list of ints
            New sequence of slots, i.e. the new slots
            will be `self.slots[indices]`.
        """
        # sanity checks
        if sorted(indices) != list(range(len(self.slots))):
            raise ValueError("Cannot reorder slots with inconclusive "
                             "ordering sequence '{}'!".format(indices))
        new_slots = []
        for idx in indices:
            new_slots.append(self.slots[idx])
        self.slots = new_slots

    def reset(self):
        """Reset the pipeline"""
        #: Filters are instances of :class:`shapeout2.pipeline.Filter`
        self.filters = []
        #: Plots are instances of :class:`shapeout2.pipeline.Plot`
        self.plots = []
        #: Filter rays of the current pipeline
        self.rays = {}
        #: Slots are instances of :class:`shapeout2.pipeline.Dataslot`
        self.slots = []
        #: individual element states
        self.element_states = {}

    def set_element_active(self, slot_id, filt_plot_id, active=True):
        """Activate an element in the block matrix"""
        self.element_states[slot_id][filt_plot_id] = active



import dclab
import numpy as np
from PyQt5 import QtCore
import pyqtgraph as pg
from pyqtgraph.graphicsItems.GradientEditorItem import Gradients

from ... import plot_cache

from .. import pipeline_plot
from ..widgets import SimplePlotWidget, SimpleViewBox


class QuickViewScatterWidget(SimplePlotWidget):
    update_hover_pos = QtCore.pyqtSignal(QtCore.QPointF)

    def __init__(self, *args, **kwargs):
        self._view_box = QuickViewViewBox()
        super(QuickViewScatterWidget, self).__init__(viewBox=self._view_box,
                                                     *args, **kwargs)
        self._view_box.update_hover_pos.connect(self.update_hover_pos)
        self.scatter = RTDCScatterPlot()
        self.select = pg.PlotDataItem(x=[1], y=[2], symbol="o")
        #: List of isoelasticity line plots
        self.isoelastics = []
        self.addItem(self.scatter)
        self.addItem(self.select)
        self.select.hide()
        self.xscale = "linear"
        self.yscale = "linear"
        #: Boolean array identifying the plotted events w.r.t. the full
        #: dataset
        self.events_plotted = None
        #: Unfiltered and not-downsampled x component of current scatter plot
        self.data_x = None
        #: Unfiltered and not-downsampled y component of current scatter plot
        self.data_y = None

        # polygon editing ROI
        self.poly_line_roi = None

        # Signals for mouse click
        # let view box update the selected event in the scatter plot
        self._view_box.set_scatter_point.connect(self.scatter.set_point)
        # let view box update self.poly_line_roi
        self._view_box.add_poly_vertex.connect(self.add_poly_vertex)

    def activate_poly_mode(self, points=[]):
        if self.poly_line_roi is None:
            self.poly_line_roi = pg.PolyLineROI([], closed=True)
            self.poly_line_roi.setPen("k")
        self.addItem(self.poly_line_roi)
        if len(points):
            mode = "poly-modify"
            self.set_poly_points(points)
        else:
            mode = "poly-create"
        self.set_mouse_click_mode(mode)
        return self.poly_line_roi

    def activate_scatter_mode(self):
        if self.poly_line_roi is not None:
            self.removeItem(self.poly_line_roi)
            self.poly_line_roi = None
        self.set_mouse_click_mode("scatter")

    def add_poly_vertex(self, pos):
        state = self.poly_line_roi.getState()
        state["points"].append([pos.x(), pos.y()])
        self.poly_line_roi.setState(state)

    def get_poly_points(self):
        if self.poly_line_roi is None:
            raise ValueError("No polygon selection active!")
        state = self.poly_line_roi.getState()
        points = np.array([[p.x(), p.y()] for p in state["points"]])
        if self.xscale == "log":
            points[:, 0] = 10**points[:, 0]
        if self.yscale == "log":
            points[:, 1] = 10**points[:, 1]
        return points

    def plot_data(self, rtdc_ds, slot, xax="area_um", yax="deform",
                  xscale="linear", yscale="linear",  downsample=False,
                  isoelastics=False):
        self.rtdc_ds = rtdc_ds
        self.slot = slot
        self.xax = xax
        self.yax = yax
        self.xscale = xscale
        self.yscale = yscale
        x, y, kde, idx = plot_cache.get_scatter_data(
            rtdc_ds=self.rtdc_ds,
            downsample=downsample,
            xax=self.xax,
            yax=self.yax,
            xscale=self.xscale,
            yscale=self.yscale)
        self.events_plotted = idx
        #: unfiltered x data
        self.data_x = self.rtdc_ds[self.xax]
        #: unfiltered y data
        self.data_y = self.rtdc_ds[self.yax]
        # define colormap
        brush = []
        cmap = pg.ColorMap(*zip(*Gradients["viridis"]["ticks"]))
        for k in kde:
            brush.append(cmap.mapToQColor(k))
        if x.size:  # test for empty x/y (#37)
            # set viewbox
            pipeline_plot.set_viewbox(plot=self,
                                      range_x=(x.min(), x.max()),
                                      range_y=(y.min(), y.max()),
                                      scale_x=self.xscale,
                                      scale_y=self.yscale,
                                      padding=0.05)
        # set data
        self.setData(x, y, brush=brush)
        # set axes labels (replace with user-defined flourescence names)
        left = dclab.dfn.get_feature_label(self.yax)
        bottom = dclab.dfn.get_feature_label(self.xax)
        for key in self.slot.fl_name_dict:
            if key in left:
                left = left.replace(key, self.slot.fl_name_dict[key])
                break
        for key in self.slot.fl_name_dict:
            if key in bottom:
                bottom = bottom.replace(key, self.slot.fl_name_dict[key])
                break
        self.plotItem.setLabels(left=left, bottom=bottom)

        # Force updating the plot item size, otherwise axes labels
        # may have an offset.
        s = self.plotItem.size()
        self.plotItem.resize(s.width()+1, s.height())
        self.plotItem.resize(s)

        # Isoelastics
        # remove old isoelastics
        for lp in self.isoelastics:
            self.removeItem(lp)
        if isoelastics:
            cfg = self.rtdc_ds.config
            self.isoelastics = pipeline_plot.add_isoelastics(
                plot_item=self.plotItem,
                axis_x=self.xax,
                axis_y=self.yax,
                channel_width=cfg["setup"]["channel width"],
                pixel_size=cfg["imaging"]["pixel size"])

    def set_mouse_click_mode(self, mode):
        allowed = ["scatter", "poly-create", "poly-modify"]
        if mode not in allowed:
            raise ValueError("Invalid mouse mode: {}, ".format(mode)
                             + "expected one of {}".format(allowed))
        if mode in ["poly-create", "poly-modify"]:
            if self.poly_line_roi is None:
                raise ValueError("Please set self.poly_line_roi before "
                                 + "setting the click mode!")
        self._view_box.mode = mode

    def set_poly_points(self, points):
        if self.poly_line_roi is None:
            raise ValueError("No polygon selection active!")
        points = np.array(points, copy=True)
        if points.size:
            if self.xscale == "log":
                points[:, 0] = np.log10(points[:, 0])
            if self.yscale == "log":
                points[:, 1] = np.log10(points[:, 1])
            state = self.poly_line_roi.getState()
            state["points"] = points.tolist()
            self.poly_line_roi.setState(state)

    def setData(self, x, y, **kwargs):
        # convert to log-scale if applicable
        if self.xscale == "log":
            x = np.log10(x)
        if self.yscale == "log":
            y = np.log10(y)
        # set data
        self.scatter.setData(x=x, y=y, **kwargs)

    def setSelection(self, event_index):
        x = self.data_x[event_index]
        y = self.data_y[event_index]
        self.select.setData([x], [y])


class QuickViewViewBox(SimpleViewBox):
    set_scatter_point = QtCore.pyqtSignal(QtCore.QPointF)
    add_poly_vertex = QtCore.pyqtSignal(QtCore.QPointF)
    update_hover_pos = QtCore.pyqtSignal(QtCore.QPointF)

    def __init__(self, *args, **kwds):
        super(QuickViewViewBox, self).__init__(*args, **kwds)
        self.mode = "scatter"

        #: allowed right-click menu options with new name
        self.right_click_actions["View All"] = "View All Content"
        self.right_click_actions["Mouse Mode"] = "Change Mouse mode"

    def mouseClickEvent(self, ev):
        if ev.button() == QtCore.Qt.LeftButton:
            pos = self.mapToView(ev.pos())
            if self.mode == "scatter":
                self.set_scatter_point.emit(pos)
            elif self.mode == "poly-create":
                self.add_poly_vertex.emit(pos)
            ev.accept()
        else:
            # right mouse button shows menu
            super(QuickViewViewBox, self).mouseClickEvent(ev)

    def hoverEvent(self, ev):
        if hasattr(ev, "_scenePos"):
            pos = self.mapToView(ev.pos())
            self.update_hover_pos.emit(pos)


class RTDCScatterPlot(pg.ScatterPlotItem):
    def __init__(self, size=3, pen=pg.mkPen(color=(0, 0, 0, 0)),
                 brush=pg.mkBrush("k"),
                 *args, **kwargs):
        super(RTDCScatterPlot, self).__init__(size=size,
                                              pen=pen,
                                              brush=brush,
                                              symbol="s",
                                              *args,
                                              **kwargs)
        self.setData(x=range(10), y=range(10))

    def pointAt(self, pos):
        """Unlike `ScatterPlotItem.pointsAt`, return the closest point"""
        x = pos.x()
        y = pos.y()
        pw = self.pixelWidth()
        ph = self.pixelHeight()
        p = self.points()[0]
        d = np.inf
        for s in self.points():
            sp = s.pos()
            di = ((sp.x() - x)/pw)**2 + ((sp.y() - y)/ph)**2
            if di < d:
                p = s
                d = di
        return p

    def set_point(self, view_pos):
        pos = self.mapFromView(view_pos)
        pt = self.pointAt(pos)
        self.ptClicked = pt
        self.sigClicked.emit(self, self.ptClicked)

    def mouseClickEvent(self, ev):
        """Override that does not handle events"""
        ev.ignore()  # clicks are handles by CustomViewBox



import copy
import pkg_resources
import warnings

import dclab
from dclab.features.emodulus.viscosity import KNOWN_MEDIA
import numpy as np
from PyQt5 import uic, QtCore, QtWidgets

from ... import meta_tool
from ...pipeline import Dataslot

from .dlg_slot_reorder import DlgSlotReorder


class SlotPanel(QtWidgets.QWidget):
    #: Emitted when a shapeout2.pipeline.Dataslot is to be changed
    slot_changed = QtCore.pyqtSignal(dict)
    #: Emitted when the pipeline is to be changed
    pipeline_changed = QtCore.pyqtSignal(dict)

    def __init__(self, *args, **kwargs):
        QtWidgets.QWidget.__init__(self)
        path_ui = pkg_resources.resource_filename(
            "shapeout2.gui.analysis", "ana_slot.ui")
        uic.loadUi(path_ui, self)
        # current Shape-Out 2 pipeline
        self._pipeline = None
        # signals
        self.toolButton_reorder.clicked.connect(self.on_reorder_slots)
        self.toolButton_anew.clicked.connect(self.on_anew_slot)
        self.toolButton_duplicate.clicked.connect(self.on_duplicate_slot)
        self.toolButton_remove.clicked.connect(self.on_remove_slot)
        self.pushButton_apply.clicked.connect(self.write_slot)
        self.pushButton_reset.clicked.connect(self.update_content)
        self.comboBox_slots.currentIndexChanged.connect(self.update_content)
        self.comboBox_medium.currentIndexChanged.connect(self.on_ui_changed)
        self.comboBox_temp.currentIndexChanged.connect(self.on_ui_changed)
        self.doubleSpinBox_temp.valueChanged.connect(self.on_ui_changed)
        # init
        self._update_emodulus_medium_choices()
        self._update_emodulus_temp_choices()

        self.update_content()

    def __getstate__(self):
        slot_state = self.current_slot_state
        if self.comboBox_temp.currentData() in ["manual", "config"]:
            emod_temp = self.doubleSpinBox_temp.value()
        else:
            emod_temp = np.nan
        if self.comboBox_medium.currentData() in KNOWN_MEDIA:
            emod_visc = np.nan  # viscosity computed from LUT
            scenario = self.comboBox_temp.currentData()
        elif self.comboBox_medium.currentData() == "unknown":
            emod_visc = np.nan  # viscosity not defined
            scenario = None
        else:  # "other", user-defined medium
            emod_visc = self.doubleSpinBox_visc.value()  # user input
            scenario = None
        state = {
            "identifier": slot_state["identifier"],
            "name": self.lineEdit_name.text(),
            "path": slot_state["path"],
            "color": self.pushButton_color.color().name(),
            "slot used": self.checkBox_use.isChecked(),
            "fl names": {"FL-1": self.lineEdit_fl1.text(),
                         "FL-2": self.lineEdit_fl2.text(),
                         "FL-3": self.lineEdit_fl3.text(),
                         },
            "crosstalk": {
                "crosstalk fl12": self.doubleSpinBox_ct12.value(),
                "crosstalk fl13": self.doubleSpinBox_ct13.value(),
                "crosstalk fl21": self.doubleSpinBox_ct21.value(),
                "crosstalk fl23": self.doubleSpinBox_ct23.value(),
                "crosstalk fl31": self.doubleSpinBox_ct31.value(),
                "crosstalk fl32": self.doubleSpinBox_ct32.value(),
            },
            "emodulus": {
                "emodulus enabled": slot_state["emodulus"]["emodulus enabled"],
                "emodulus model": "elastic sphere",
                # It is ok if we have user-defined strings here, because
                # only media in KNOWN_MEDIA are passed to dclab in the end.
                "emodulus medium": self.comboBox_medium.currentData(),
                "emodulus scenario": scenario,
                "emodulus temperature": emod_temp,
                "emodulus viscosity": emod_visc,
            }
        }
        return state

    def __setstate__(self, state):
        cur_state = self.current_slot_state
        if cur_state["identifier"] != state["identifier"]:
            raise ValueError("Slot identifier mismatch!")
        self.lineEdit_name.setText(state["name"])
        self.lineEdit_path.setText(str(state["path"]))
        self.pushButton_color.setColor(state["color"])
        self.lineEdit_fl1.setText(state["fl names"]["FL-1"])
        self.lineEdit_fl2.setText(state["fl names"]["FL-2"])
        self.lineEdit_fl3.setText(state["fl names"]["FL-3"])
        self.checkBox_use.setChecked(state["slot used"])
        # crosstalk
        crosstalk = state["crosstalk"]
        self.doubleSpinBox_ct12.setValue(crosstalk["crosstalk fl12"])
        self.doubleSpinBox_ct13.setValue(crosstalk["crosstalk fl13"])
        self.doubleSpinBox_ct21.setValue(crosstalk["crosstalk fl21"])
        self.doubleSpinBox_ct23.setValue(crosstalk["crosstalk fl23"])
        self.doubleSpinBox_ct31.setValue(crosstalk["crosstalk fl31"])
        self.doubleSpinBox_ct32.setValue(crosstalk["crosstalk fl32"])
        # emodulus
        # updating the medium/temperature choices has to be done first,
        # because self.comboBox_medium triggers the function on_ui_changed.
        self._update_emodulus_medium_choices()
        self._update_emodulus_temp_choices()
        emodulus = state["emodulus"]
        self.groupBox_emod.setVisible(emodulus["emodulus enabled"])
        idx_med = self.comboBox_medium.findData(emodulus["emodulus medium"])
        if idx_med == -1:  # empty medium string
            idx_med = self.comboBox_medium.findData("other")
        self.comboBox_medium.setCurrentIndex(idx_med)
        cfg = meta_tool.get_rtdc_config(state["path"])
        if "medium" in cfg["setup"] and cfg["setup"]["medium"] in KNOWN_MEDIA:
            self.comboBox_medium.setEnabled(False)  # prevent modification
        else:
            self.comboBox_medium.setEnabled(True)  # user-defined
        # https://dclab.readthedocs.io/en/latest/sec_av_emodulus.html
        scenario = emodulus.get("emodulus scenario", "manual")
        if scenario:
            idx_scen = self.comboBox_temp.findData(scenario)
            self.comboBox_temp.blockSignals(True)
            self.comboBox_temp.setCurrentIndex(idx_scen)
            self.comboBox_temp.blockSignals(False)
        # This has to be done after setting the scenario
        # (otherwise it might be overridden in the frontend)
        self.doubleSpinBox_temp.setValue(emodulus["emodulus temperature"])
        self.doubleSpinBox_visc.setValue(emodulus["emodulus viscosity"])

        # Fluorescence data visibility
        features = meta_tool.get_rtdc_features(state["path"])
        hasfl1 = "fl1_max" in features
        hasfl2 = "fl2_max" in features
        hasfl3 = "fl3_max" in features

        # labels
        self.lineEdit_fl1.setVisible(hasfl1)
        self.label_fl1.setVisible(hasfl1)
        self.lineEdit_fl2.setVisible(hasfl2)
        self.label_fl2.setVisible(hasfl2)
        self.lineEdit_fl3.setVisible(hasfl3)
        self.label_fl3.setVisible(hasfl3)

        # crosstalk matrix
        self.label_from_fl1.setVisible(hasfl1 & hasfl2 | hasfl1 & hasfl3)
        self.label_from_fl2.setVisible(hasfl2 & hasfl1 | hasfl2 & hasfl3)
        self.label_from_fl3.setVisible(hasfl3 & hasfl1 | hasfl3 & hasfl2)
        self.label_to_fl1.setVisible(hasfl1 & hasfl2 | hasfl1 & hasfl3)
        self.label_to_fl2.setVisible(hasfl2 & hasfl1 | hasfl2 & hasfl3)
        self.label_to_fl3.setVisible(hasfl3 & hasfl1 | hasfl3 & hasfl2)
        self.doubleSpinBox_ct12.setVisible(hasfl1 & hasfl2)
        self.doubleSpinBox_ct13.setVisible(hasfl1 & hasfl3)
        self.doubleSpinBox_ct21.setVisible(hasfl2 & hasfl1)
        self.doubleSpinBox_ct23.setVisible(hasfl2 & hasfl3)
        self.doubleSpinBox_ct31.setVisible(hasfl3 & hasfl1)
        self.doubleSpinBox_ct32.setVisible(hasfl3 & hasfl2)

        self.groupBox_fl_labels.setVisible(hasfl1 | hasfl2 | hasfl3)
        self.groupBox_fl_cross.setVisible(hasfl1 | hasfl2 | hasfl3)

    @staticmethod
    def get_dataset_choices_medium(ds):
        """Return the choices for the medium selection

        Parameters
        ----------
        ds: RTDCBase
            Dataset

        Returns
        -------
        choices: list
            List of [title, identifier]
        """
        if ds:
            medium = ds.config.get("setup", {}).get("medium", "").strip()
            if not medium:  # empty medium string
                medium = "other"
        else:
            medium = "undefined"
        if medium in KNOWN_MEDIA:
            valid_media = [medium]
        else:
            valid_media = KNOWN_MEDIA + [medium, "other", "undefined"]
        choices = []
        for vm in valid_media:
            if vm == "CellCarrierB":
                name = "CellCarrier B"  # [sic]
            else:
                name = vm
            choices.append([name, vm])
        return choices

    @staticmethod
    def get_dataset_choices_temperature(ds):
        """Return the choices for the temperature selection

        Parameters
        ----------
        ds: RTDCBase
            Dataset

        Returns
        -------
        choices: list
            List of [title, identifier]
        """
        choices = []
        if ds is not None:
            if "temp" in ds:
                choices.append(["From feature", "feature"])
            if "temperature" in ds.config["setup"]:
                choices.append(["From meta data", "config"])
        choices.append(["Manual", "manual"])
        return choices

    def _update_emodulus_medium_choices(self):
        """update currently available medium choices for YM

        Signals are blocked.
        """
        self.comboBox_medium.blockSignals(True)
        self.comboBox_medium.clear()
        ds = self.get_dataset()
        choices = self.get_dataset_choices_medium(ds)
        for name, data in choices:
            self.comboBox_medium.addItem(name, data)
        self.comboBox_medium.blockSignals(False)

    def _update_emodulus_temp_choices(self):
        """pupdate temperature choices for YM

        The previous selection is preserved. Signals are blocked.
        """
        self.comboBox_temp.blockSignals(True)
        cursel = self.comboBox_temp.currentData()
        self.comboBox_temp.clear()
        ds = self.get_dataset()
        choices = self.get_dataset_choices_temperature(ds)
        for name, data in choices:
            self.comboBox_temp.addItem(name, data)
        idx = self.comboBox_temp.findData(cursel)
        self.comboBox_temp.setCurrentIndex(idx)
        self.comboBox_temp.blockSignals(False)

    @property
    def current_slot_state(self):
        if self.slot_ids:
            slot_index = self.comboBox_slots.currentIndex()
            slot_state = self.pipeline.slots[slot_index].__getstate__()
        else:
            slot_state = None
        return slot_state

    @property
    def pipeline(self):
        return self._pipeline

    @property
    def slot_ids(self):
        """List of slot identifiers"""
        if self.pipeline is None:
            return []
        else:
            return [slot.identifier for slot in self.pipeline.slots]

    @property
    def slot_names(self):
        """List of slot names"""
        if self.pipeline is None:
            return []
        else:
            return [slot.name for slot in self.pipeline.slots]

    def get_dataset(self):
        """Return dataset associated with the current slot index

        Returns None if there is no dataset in the pipeline.
        """
        if self.pipeline is not None and self.pipeline.slots:
            slot_index = self.comboBox_slots.currentIndex()
            slot = self.pipeline.slots[slot_index]
            return slot.get_dataset()
        else:
            return None

    @QtCore.pyqtSlot()
    def on_anew_slot(self):
        slot_state = self.__getstate__()
        new_slot = Dataslot(slot_state["path"])
        pos = self.pipeline.slot_ids.index(slot_state["identifier"])
        self.pipeline.add_slot(new_slot, index=pos+1)
        state = self.pipeline.__getstate__()
        self.pipeline_changed.emit(state)

    @QtCore.pyqtSlot()
    def on_duplicate_slot(self):
        # determine the new filter state
        slot_state = self.__getstate__()
        new_state = copy.deepcopy(slot_state)
        new_slot = Dataslot(slot_state["path"])
        new_state["identifier"] = new_slot.identifier
        new_state["name"] = new_slot.name
        new_slot.__setstate__(new_state)
        # determine the filter position
        pos = self.pipeline.slot_ids.index(slot_state["identifier"])
        self.pipeline.add_slot(new_slot, index=pos+1)
        state = self.pipeline.__getstate__()
        self.pipeline_changed.emit(state)

    @QtCore.pyqtSlot()
    def on_remove_slot(self):
        slot_state = self.__getstate__()
        self.pipeline.remove_slot(slot_state["identifier"])
        state = self.pipeline.__getstate__()
        self.pipeline_changed.emit(state)

    @QtCore.pyqtSlot()
    def on_reorder_slots(self):
        """Open dialog for reordering slots"""
        dlg = DlgSlotReorder(self.pipeline, self)
        dlg.pipeline_changed.connect(self.pipeline_changed)
        dlg.exec()

    @QtCore.pyqtSlot()
    def on_ui_changed(self):
        """Called when the user modifies the medium or temperature options"""
        medium = self.comboBox_medium.currentData()
        tselec = self.comboBox_temp.currentData()
        if medium in KNOWN_MEDIA:  # medium registered with dclab
            self.comboBox_temp.setEnabled(True)
            self.doubleSpinBox_visc.setEnabled(True)
            self.doubleSpinBox_visc.setReadOnly(True)
            if tselec == "manual":
                temperature = self.doubleSpinBox_temp.value()
                self.doubleSpinBox_temp.setEnabled(True)
                self.doubleSpinBox_temp.setReadOnly(False)
            elif tselec == "config":
                # get temperature from dataset
                ds = self.get_dataset()
                temperature = ds.config["setup"]["temperature"]
                self.doubleSpinBox_temp.setEnabled(True)
                self.doubleSpinBox_temp.setReadOnly(True)
                self.doubleSpinBox_temp.setValue(temperature)
            elif tselec == "feature":
                temperature = np.nan
                self.doubleSpinBox_temp.setEnabled(False)
                self.doubleSpinBox_temp.setValue(temperature)
            else:
                assert tselec is None, "We should still be in init"
                return
            # For user convenience, also show the viscosity
            if (medium in dclab.features.emodulus.viscosity.KNOWN_MEDIA
                    and not np.isnan(temperature)):
                # compute viscosity
                state = self.__getstate__()
                cfg = meta_tool.get_rtdc_config(state["path"])
                with warnings.catch_warnings(record=True) as w:
                    # Warn the user if the temperature is out-of-range
                    warnings.simplefilter("always")
                    visc = dclab.features.emodulus.viscosity.get_viscosity(
                        medium=medium,
                        channel_width=cfg["setup"]["channel width"],
                        flow_rate=cfg["setup"]["flow rate"],
                        temperature=temperature)
                    for wi in w:
                        if issubclass(wi.category,
                                      dclab.features.emodulus.viscosity.
                                      TemperatureOutOfRangeWarning):
                            vstyle = "color: #950000; border-width: 2px"
                            break
                    else:
                        vstyle = "border-width: 2px"
                self.doubleSpinBox_visc.setEnabled(True)
                self.doubleSpinBox_visc.setReadOnly(True)
                self.doubleSpinBox_visc.setValue(visc)
                self.doubleSpinBox_visc.setStyleSheet(vstyle)
            else:
                self.doubleSpinBox_visc.setEnabled(False)
                self.doubleSpinBox_visc.setReadOnly(True)
                self.doubleSpinBox_visc.setValue(np.nan)
                self.doubleSpinBox_visc.setStyleSheet("border-width: 2px")
        elif medium == "undefined":
            self.comboBox_temp.setEnabled(False)
            self.doubleSpinBox_temp.setEnabled(False)
            self.doubleSpinBox_temp.setValue(np.nan)
            self.doubleSpinBox_visc.setValue(np.nan)
            self.doubleSpinBox_visc.setEnabled(False)
            self.doubleSpinBox_visc.setStyleSheet("border-width: 2px")
        else:  # "other" or user-defined
            self.comboBox_temp.setEnabled(False)
            self.doubleSpinBox_temp.setEnabled(False)
            self.doubleSpinBox_temp.setValue(np.nan)
            self.doubleSpinBox_visc.setEnabled(True)
            self.doubleSpinBox_visc.setReadOnly(False)
            self.doubleSpinBox_visc.setStyleSheet("border-width: 2px")

    def set_pipeline(self, pipeline):
        self._pipeline = pipeline

    def show_slot(self, slot_id):
        self.update_content(slot_index=self.slot_ids.index(slot_id))

    def update_content(self, event=None, slot_index=None):
        if self.slot_ids:
            self.setEnabled(True)
            # update combobox
            self.comboBox_slots.blockSignals(True)
            if slot_index is None:
                slot_index = self.comboBox_slots.currentIndex()
                if slot_index > len(self.slot_ids) - 1 or slot_index < 0:
                    slot_index = len(self.slot_ids) - 1
            self.comboBox_slots.clear()
            self.comboBox_slots.addItems(self.slot_names)
            self.comboBox_slots.setCurrentIndex(slot_index)
            self.comboBox_slots.blockSignals(False)
            # populate content
            slot_state = self.pipeline.slots[slot_index].__getstate__()
            self.__setstate__(slot_state)
            self.on_ui_changed()
        else:
            self.setEnabled(False)

    def write_slot(self):
        """Update the shapeout2.pipeline.Dataslot instance"""
        # get current index
        slot_state = self.__getstate__()
        slot = self.pipeline.get_slot(slot_state["identifier"])
        # This is important, otherwise update_content will not have the
        # latest state.
        slot.__setstate__(slot_state)
        self.update_content()  # update slot combobox and visible fl names
        self.slot_changed.emit(slot_state)



from PyQt5 import QtCore, QtWidgets
import pyqtgraph as pg
from pyqtgraph import exporters


class SimplePlotItem(pg.PlotItem):
    """Custom class for data visualization in Shape-Out

    Modifications include:
    - right click menu only with "Export..."
    - top and right axes
    """

    def __init__(self, parent=None, *args, **kwargs):
        if "viewBox" not in kwargs:
            kwargs["viewBox"] = SimpleViewBox()
        super(SimplePlotItem, self).__init__(parent, *args, **kwargs)
        self.vb.export.connect(self.on_export)
        # show top and right axes, but not ticklabels
        for kax in ["top", "right"]:
            self.showAxis(kax)
            ax = self.axes[kax]["item"]
            ax.setTicks([])
            ax.setLabel(None)
            ax.setStyle(tickTextOffset=0,
                        tickTextWidth=0,
                        tickTextHeight=0,
                        autoExpandTextSpace=False,
                        showValues=False,
                        )

        # show grid
        self.showGrid(x=True, y=True, alpha=.1)
        # visualization
        self.hideButtons()

    def axes_to_front(self):
        """Give the axes a high zValue"""
        # bring axes to front
        # (This screws up event selection in QuickView)
        for kax in self.axes:
            self.axes[kax]["item"].setZValue(200)

    def on_export(self, suffix):
        """Export subplots as original figures (with axes labels, etc)"""
        file, _ = QtWidgets.QFileDialog.getSaveFileName(
            None,
            'Save {} file'.format(suffix.upper()),
            '',
            '{} file (*.{})'.format(suffix.upper(), suffix))
        if not file.endswith("." + suffix):
            file += "." + suffix
        self.perform_export(file)

    def perform_export(self, file):
        suffix = file[-3:]
        if suffix == "png":
            exp = exporters.ImageExporter(self)
            # translate from screen resolution (80dpi) to 300dpi
            exp.params["width"] = int(exp.params["width"] / 72 * 300)
        elif suffix == "svg":
            exp = exporters.SVGExporter(self)
        exp.export(file)


class SimplePlotWidget(pg.PlotWidget):
    """Custom class for data visualization in Shape-Out

    Modifications include:
    - white background
    - those of SimplePlotItem
    """

    def __init__(self, parent=None, background='w', **kargs):
        # The following code is copied from pg.PlotWidget and instead
        # of PlotItem we use SimplePlotItem.
        pg.GraphicsView.__init__(self, parent, background=background)
        self.enableMouse(False)
        self.plotItem = SimplePlotItem(**kargs)
        self.setCentralItem(self.plotItem)
        # Explicitly wrap methods from plotItem
        for m in [
            'addItem', 'removeItem', 'autoRange', 'clear', 'setXRange',
            'setYRange', 'setRange', 'setAspectLocked', 'setMouseEnabled',
            'setXLink', 'setYLink', 'enableAutoRange', 'disableAutoRange',
                'setLimits', 'register', 'unregister', 'viewRect']:
            setattr(self, m, getattr(self.plotItem, m))
        self.plotItem.sigRangeChanged.connect(self.viewRangeChanged)


class SimpleViewBox(pg.ViewBox):
    export = QtCore.pyqtSignal(str)

    def __init__(self, *args, **kwargs):
        super(SimpleViewBox, self).__init__(*args, **kwargs)
        #: allowed right-click menu options with new name
        self.right_click_actions = {}
        settings = QtCore.QSettings()
        if int(settings.value("advanced/developer mode", 0)):
            # Enable advanced export in developer mode
            self.right_click_actions["Export..."] = "Advanced Export"

    def raiseContextMenu(self, ev):
        # Let the scene add on to the end of our context menu
        menu = self.scene().addParentContextMenus(self, self.menu, ev)

        # Only keep list of actions defined in `self.right_click_actions`
        for action in self.menu.actions():
            if action.text() in self.right_click_actions.values():
                pass
            elif action.text() not in self.right_click_actions:
                self.menu.removeAction(action)
            else:
                action.setText(self.right_click_actions[action.text()])

        menu.addAction("Export subplot as PNG",
                       lambda: self.export.emit("png"))
        menu.addAction("Export subplot as SVG",
                       lambda: self.export.emit("svg"))

        pos = ev.screenPos()
        menu.popup(QtCore.QPoint(pos.x(), pos.y()))
        return True



import copy

import numpy as np
from PyQt5 import QtCore, QtWidgets

from ... import pipeline

from .dm_dataset import MatrixDataset
from .dm_filter import MatrixFilter
from .dm_element import MatrixElement


class DataMatrix(QtWidgets.QWidget):
    quickviewed = QtCore.pyqtSignal(int, int)
    matrix_changed = QtCore.pyqtSignal()
    filter_modify_clicked = QtCore.pyqtSignal(str)
    slot_modify_clicked = QtCore.pyqtSignal(str)

    def __init__(self, parent=None):
        super(DataMatrix, self).__init__(parent)

        self.glo = None
        self._reset_layout()

        self.setAcceptDrops(True)

        # used for toggling between all active, all inactive and semi state
        self.semi_states_dataset = {}
        self.semi_states_filter = {}

        # used for remembering quickview element
        self._old_quickview_instance = None

    def __getstate__(self):
        """State of the current data matrix"""
        # slots
        slot_states = []
        slots_used = []
        for dw in self.dataset_widgets:
            dw_state = dw.__getstate__()
            slot = pipeline.Dataslot.get_slot(dw_state["identifier"])
            slot.slot_used = dw_state["enabled"]
            slot_states.append(slot.__getstate__())
            if dw_state["enabled"]:
                slots_used.append(dw_state["identifier"])

        # filters
        filter_states = []
        filters_used = []
        for fw in self.filter_widgets:
            fw_state = fw.__getstate__()
            filt = pipeline.Filter.get_filter(fw_state["identifier"])
            filter_states.append(filt.__getstate__())
            if fw_state["enabled"]:
                filters_used.append(fw_state["identifier"])
        # elements
        mestates = {}
        for dw in self.dataset_widgets:
            idict = {}
            for fw in self.filter_widgets:
                me = self.get_matrix_element(dw.identifier, fw.identifier)
                # We only store the information about whether the user
                # clicked this element. The state about "enabled" is stored
                # in `slots_used` and `filters_used`.
                idict[fw.identifier] = me.__getstate__()["active"]
            mestates[dw.identifier] = idict
        state = {"elements": mestates,
                 "filters": filter_states,
                 "filters used": filters_used,
                 "slots": slot_states,
                 "slots used": slots_used,
                 }
        return state

    def __setstate__(self, state):
        # remember current QuickView identifiers
        qv_slot_id, qv_filt_id = self.get_quickview_ids()
        self.setUpdatesEnabled(False)
        self.blockSignals(True)
        self.clear()
        # dataset states
        for ii in range(len(state["slots"])):
            slot_id = state["slots"][ii]["identifier"]
            dw_state = {"path": state["slots"][ii]["path"],
                        "identifier": slot_id,
                        "enabled": slot_id in state["slots used"],
                        }
            self.add_dataset(state=dw_state)
        # filter states
        for jj in range(len(state["filters"])):
            filt_id = state["filters"][jj]["identifier"]
            fw_state = {"identifier": filt_id,
                        "enabled": filt_id in state["filters used"],
                        "name": state["filters"][jj]["name"]
                        }
            self.add_filter(state=fw_state)
        # make sure elements exist
        # (this also sets enabled/disabled state)
        self.fill_elements()
        # element states
        MatrixElement._quick_view_instance = None
        for slot_id in state["elements"]:
            ds_state = state["elements"][slot_id]
            for filt_id in ds_state:
                me = self.get_matrix_element(slot_id, filt_id)
                me_state = me.__getstate__()
                me_state["active"] = ds_state[filt_id]
                me.__setstate__(me_state)
        # re-apply current quickview ids
        try:
            meqv = self.get_matrix_element(qv_slot_id, qv_filt_id)
        except KeyError:
            pass
        else:
            MatrixElement._quick_view_instance = meqv
            self.update_content()
        self.blockSignals(False)
        self.setUpdatesEnabled(True)

    def _reset_layout(self):
        if self.glo is not None:
            # send old layout to Nirvana eventually
            layout = self.glo
            self.glo = None
            self.old_layout = QtWidgets.QWidget()
            self.old_layout.setLayout(layout)
            self.old_layout.hide()
            self.old_layout.deleteLater()
        # add new layout
        self.glo = QtWidgets.QGridLayout()
        self.glo.setSpacing(2)
        self.glo.setContentsMargins(0, 0, 0, 0)
        # add dummy corner element
        cl = QtWidgets.QLabel("Block\nMatrix")
        cl.setAlignment(QtCore.Qt.AlignCenter)
        self.glo.addWidget(cl, 0, 0)
        self.setLayout(self.glo)

    @property
    def dataset_widgets(self):
        datasets = []
        for ii in range(self.glo.rowCount()):
            item = self.glo.itemAtPosition(ii+1, 0)
            if item is not None:
                ds = item.widget()
                datasets.append(ds)
        return datasets

    @property
    def element_width(self):
        """Data matrix element width (without 2px spacing)"""
        for jj in range(1, self.glo.columnCount()-1):
            item = self.glo.itemAtPosition(0, jj)
            if item is not None:
                width = item.geometry().width()
                break
        else:
            width = 67
        return width

    @property
    def element_widget_dict(self):
        els = {}
        for ii, ws in enumerate(self.dataset_widgets):
            elsd = {}
            for jj, wf in enumerate(self.filter_widgets):
                it = self.glo.itemAtPosition(ii+1, jj+1)
                elsd[wf.identifier] = it.widget()
            els[ws.identifier] = elsd
        return els

    @property
    def filter_widgets(self):
        filters = []
        for jj in range(self.glo.columnCount()):
            item = self.glo.itemAtPosition(0, jj+1)
            if item is not None:
                fs = item.widget()
                filters.append(fs)
        return filters

    @property
    def num_datasets(self):
        count = 0
        for ii in range(1, self.glo.rowCount()):
            if self.glo.itemAtPosition(ii, 0) is not None:
                count += 1
        return count

    @property
    def num_filters(self):
        count = 0
        for jj in range(1, self.glo.columnCount()):
            if self.glo.itemAtPosition(0, jj) is not None:
                count += 1
        return count

    @property
    def plot_matrix(self):
        for ch in self.parent().children():
            if ch.__class__.__name__ == "PlotMatrix":
                break
        else:
            raise KeyError("PlotMatrix not found!")
        return ch

    def add_dataset(self, slot_id=None, state=None):
        """Add a dataset to the DataMatrix"""
        self.setUpdatesEnabled(False)
        md = MatrixDataset(identifier=slot_id, state=state)
        self.glo.addWidget(md, self.num_datasets+1, 0)
        md.active_toggled.connect(self.toggle_dataset_active)
        md.enabled_toggled.connect(self.toggle_dataset_enable)
        md.enabled_toggled.connect(self.plot_matrix.toggle_dataset_enable)
        md.option_action.connect(self.on_option_dataset)
        md.modify_clicked.connect(self.slot_modify_clicked.emit)
        self.fill_elements()
        self.plot_matrix.fill_elements()
        self.adjust_size()
        self.plot_matrix.adjust_size()  # important when opt/removing slots
        self.setUpdatesEnabled(True)
        self.publish_matrix()
        return md

    def add_filter(self, identifier=None, state=None):
        self.setUpdatesEnabled(False)
        mf = MatrixFilter(identifier=identifier, state=state)
        mf.active_toggled.connect(self.toggle_filter_active)
        mf.enabled_toggled.connect(self.toggle_filter_enable)
        mf.option_action.connect(self.on_option_filter)
        mf.modify_clicked.connect(self.filter_modify_clicked.emit)
        self.glo.addWidget(mf, 0, self.num_filters+1)
        self.fill_elements()
        self.adjust_size()
        self.plot_matrix.adjust_size()  # important when opt/removing filters
        self.setUpdatesEnabled(True)
        self.publish_matrix()
        return mf

    def adjust_size(self):
        QtWidgets.QApplication.processEvents()
        self.setMinimumSize(self.sizeHint())
        self.setFixedSize(self.sizeHint())
        QtWidgets.QApplication.processEvents()
        self.setMinimumSize(self.sizeHint())
        self.setFixedSize(self.sizeHint())

    def changed_element(self):
        self.publish_matrix()

    def changed_quickview(self):
        slot_index_qv, filt_index_qv = self.get_quickview_indices()
        if slot_index_qv is not None and filt_index_qv is not None:
            self.quickviewed.emit(slot_index_qv, filt_index_qv)

    def enable_quickview(self, b=True):
        if b:
            MatrixElement._quick_view_instance = self._old_quickview_instance
        else:
            self._old_quickview_instance = MatrixElement._quick_view_instance
            MatrixElement._quick_view_instance = None
        self.update_content()

    def clear(self):
        """Reset layout"""
        self._reset_layout()

    def dragEnterEvent(self, event):
        print("drag enter event on data matrix")
        event.ignore()

    def dropEvent(self, event):
        print("drag drop event on data matrix")
        event.ignore()

    def fill_elements(self):
        # add widgets
        for ii in range(self.num_datasets):
            for jj in range(self.num_filters):
                if self.glo.itemAtPosition(ii+1, jj+1) is None:
                    me = MatrixElement()
                    me.quickview_selected.connect(self.changed_quickview)
                    me.element_changed.connect(self.changed_element)
                    self.glo.addWidget(me, ii+1, jj+1)
        # make sure enabled/disabled is honored
        state = self.__getstate__()
        for slot_sate in state["slots"]:
            slot_id = slot_sate["identifier"]
            for filt_state in state["filters"]:
                filt_id = filt_state["identifier"]
                if not (slot_id in state["slots used"]
                        and filt_id in state["filters used"]):
                    me = self.get_matrix_element(slot_id, filt_id)
                    mstate = me.__getstate__()
                    mstate["enabled"] = False
                    me.__setstate__(mstate)

    def get_filter_index(self, filter_id):
        for ii, fs in enumerate(self.filter_widgets):
            if fs.identifier == filter_id:
                break
        else:
            raise KeyError("Filter '{}' not found!".format(filter_id))
        return ii

    def get_filter_widget_state(self, filter_id):
        ii = self.get_filter_index(filter_id)
        fw = self.filter_widgets[ii]
        return fw.__getstate__()

    def get_slot_index(self, slot_id):
        for ii, ds in enumerate(self.dataset_widgets):
            if ds.identifier == slot_id:
                break
        else:
            raise KeyError("Dataset '{}' not found!".format(slot_id))
        return ii

    def get_slot_widget_state(self, slot_id, ret_index=False):
        ii = self.get_slot_index(slot_id)
        dw = self.dataset_widgets[ii]
        if ret_index:
            return dw.__getstate__(), ii
        else:
            return dw.__getstate__()

    def get_matrix_element(self, slot_id, filt_id):
        """Return matrix element matching dataset and filter identifiers"""
        ii, jj = self.get_matrix_indices(slot_id, filt_id)
        return self.glo.itemAtPosition(ii, jj).widget()

    def get_matrix_indices(self, slot_id, filt_id):
        ncols = self.glo.columnCount()
        nrows = self.glo.rowCount()
        for ii in range(1, nrows):
            ds = self.glo.itemAtPosition(ii, 0).widget()
            if ds.identifier == slot_id:
                for jj in range(1, ncols):
                    f = self.glo.itemAtPosition(0, jj).widget()
                    if f.identifier == filt_id:
                        break
                else:
                    raise KeyError("Filter '{}' not found!".format(filt_id))
                break
        else:
            raise KeyError("Dataset '{}' not found!".format(slot_id))
        return ii, jj

    def get_quickview_ids(self):
        current = MatrixElement._quick_view_instance
        if current is not None:
            try:
                state = self.__getstate__()
            except KeyError:
                # the state is not valid (issue #25)
                return None, None
            for slot_id in state["elements"]:
                ds_state = state["elements"][slot_id]
                for filt_id in ds_state:
                    me = self.get_matrix_element(slot_id, filt_id)
                    if current == me:
                        return slot_id, filt_id
            else:
                # no valid QuickView selection (issue #38)
                return None, None
        else:
            return None, None

    def get_quickview_indices(self):
        slot_id, filt_id = self.get_quickview_ids()
        if slot_id is not None:
            ii, jj = self.get_matrix_indices(slot_id, filt_id)
            return ii - 1, jj - 1
        else:
            return None, None

    @QtCore.pyqtSlot(str)
    def on_option_dataset(self, option):
        """Dataset option logic (remove, insert_anew, duplicate)"""
        self.setUpdatesEnabled(False)
        dw_state = self.sender().__getstate__()
        slot_id = dw_state["identifier"]
        slot_index = self.get_slot_index(slot_id)
        state = self.__getstate__()
        pstate = self.plot_matrix.__getstate__()
        if option == "remove":
            state["slots"].pop(slot_index)
            if slot_id in state["slots used"]:
                state["slots used"].remove(slot_id)
            state["elements"].pop(slot_id)
            pstate["elements"].pop(slot_id)
        else:
            # create a new slot
            slot = pipeline.Dataslot(path=dw_state["path"])
            new_id = slot.identifier
            if option == "duplicate":
                orig_state = copy.deepcopy(state["slots"][slot_index])
                # use original state
                new_state = orig_state
                new_state["identifier"] = new_id
                new_state["name"] = slot.name
                new_state["color"] = slot.color
                # also set element states
                # state["elements"][new_id] = state["elements"][slot_id]
            else:  # insert_anew
                new_state = slot.__getstate__()
            slot.__setstate__(new_state)
            state["slots"].insert(slot_index+1, new_state)
            state["slots used"].append(new_id)
        # this also takes care of filling up matrix elements
        self.__setstate__(state)
        # this correctly assigns elements in plot matrix
        # (also when option is not remove)
        self.plot_matrix.__setstate__(pstate)
        self.plot_matrix.fill_elements()
        self.publish_matrix()
        self.setUpdatesEnabled(True)

    @QtCore.pyqtSlot(str)
    def on_option_filter(self, option):
        """Filter option logic (remove, duplicate)"""
        fw_state = self.sender().__getstate__()
        filt_id = fw_state["identifier"]
        filt_index = self.get_filter_index(filt_id)
        state = self.__getstate__()
        if option == "remove":
            state["filters"].pop(filt_index)
            # remove matrix elements
            for ds_key in state["elements"]:
                state["elements"][ds_key].pop(filt_id)
        else:  # duplicate
            filt = pipeline.Filter()
            new_state = copy.deepcopy(state["filters"][filt_index])
            new_state["identifier"] = filt.identifier
            new_state["name"] = filt.name
            state["filters"].insert(filt_index+1, new_state)
            state["filters used"].append(filt.identifier)
            filt.__setstate__(new_state)
        self.__setstate__(state)
        self.publish_matrix()

    def publish_matrix(self):
        """Publish state via self.matrix_changed signal for Pipeline"""
        if not self.signalsBlocked():
            self.matrix_changed.emit()
            self.changed_quickview()

    @QtCore.pyqtSlot()
    def toggle_dataset_active(self):
        """Switch between all active, all inactive, previous state

        Modifies the matrix elements for a dataset/row,
        which is defined by the signal sender :class:`MatrixDataset`.
        Cyclic toggling order: semi -> all -> none
        """
        self.semi_states_filter = {}  # sic
        sender = self.sender()
        slot_id = sender.identifier
        state = self.__getstate__()["elements"][slot_id]
        num_actives = sum([s for s in state.values()])

        # update state according to the scheme in the docstring
        if num_actives == 0:
            if slot_id in self.semi_states_dataset:
                # use semi state
                oldstate = self.semi_states_dataset[slot_id]
                for filt_id in oldstate:
                    if filt_id in state:
                        state[filt_id] = oldstate[filt_id]
            else:
                # toggle all to active
                for filt_id in state:
                    state[filt_id] = True
        elif num_actives == len(state):
            # toggle all to inactive
            for filt_id in state:
                state[filt_id] = False
        else:
            # save semi state
            self.semi_states_dataset[slot_id] = copy.deepcopy(state)
            # toggle all to active
            for filt_id in state:
                state[filt_id] = True

        for filt_id in state:
            me = self.get_matrix_element(slot_id, filt_id)
            me.set_active(state[filt_id])
        self.publish_matrix()

    @QtCore.pyqtSlot(bool)
    def toggle_dataset_enable(self, enabled):
        sender = self.sender()
        slot_id = sender.identifier
        state = self.__getstate__()
        for filt_id in state["elements"][slot_id]:
            # make sure that disabled filters are honored
            fstate = self.get_filter_widget_state(filt_id)
            fenabled = fstate["enabled"]
            # update element widget
            me = self.get_matrix_element(slot_id, filt_id)
            mstate = me.__getstate__()
            mstate["enabled"] = np.logical_and(enabled, fenabled)
            me.__setstate__(mstate)
        self.publish_matrix()

    @QtCore.pyqtSlot()
    def toggle_filter_active(self):
        """Switch between all active, all inactive, previous state

        Modifies the matrix elements for a filter/column,
        which is defined by the signal sender :class:`MatrixFilter`.
        Cyclic toggling order: semi -> all -> none
        """
        self.semi_states_dataset = {}  # sic
        sender = self.sender()
        filt_id = sender.identifier

        states = self.__getstate__()["elements"]
        state = {}
        for slot_id in states:
            state[slot_id] = states[slot_id][filt_id]

        num_actives = sum(list(state.values()))

        # update state according to the scheme in the docstring
        if num_actives == 0:
            if filt_id in self.semi_states_filter:
                # use semi state
                oldstate = self.semi_states_filter[filt_id]
                for slot_id in oldstate:
                    if slot_id in state:
                        state[slot_id] = oldstate[slot_id]
            else:
                # toggle all to active
                for slot_id in state:
                    state[slot_id] = True
        elif num_actives == len(state):
            # toggle all to inactive
            for slot_id in state:
                state[slot_id] = False
        else:
            # save semi state
            self.semi_states_filter[filt_id] = copy.deepcopy(state)
            # toggle all to active
            for slot_id in state:
                state[slot_id] = True

        for slot_id in state:
            me = self.get_matrix_element(slot_id, filt_id)
            me.set_active(state[slot_id])
        self.publish_matrix()

    @QtCore.pyqtSlot(bool)
    def toggle_filter_enable(self, enabled):
        sender = self.sender()
        sid = sender.identifier
        state = self.__getstate__()
        for slot_id in state["elements"]:
            # make sure that disabled filters are honored
            dstate = self.get_slot_widget_state(slot_id)
            denabled = dstate["enabled"]
            # update element widget
            me = self.get_matrix_element(slot_id, sid)
            mstate = me.__getstate__()
            mstate["enabled"] = np.logical_and(enabled, denabled)
            me.__setstate__(mstate)
        self.publish_matrix()

    def update_content(self):
        ncols = self.glo.columnCount()
        nrows = self.glo.rowCount()
        for ii in range(nrows):
            for jj in range(ncols):
                if ii == 0 and jj == 0:
                    # block matrix label
                    continue
                item = self.glo.itemAtPosition(ii, jj)
                if item is not None:
                    item.widget().update_content()



import pathlib
import pkg_resources

import numpy as np
from PyQt5 import QtCore, QtWidgets


class KeyValueTableWidget(QtWidgets.QTableWidget):
    def __init__(self, *args, **kwargs):
        """A table widgets with two columns for key-value visualization"""
        super(KeyValueTableWidget, self).__init__(*args, **kwargs)
        path_css = pkg_resources.resource_filename(
            "shapeout2.gui.widgets", "key_value_table_widget.css")
        stylesheet = pathlib.Path(path_css).read_text()
        self.setStyleSheet(stylesheet)
        self.horizontalHeader().hide()
        self.verticalHeader().hide()
        self.setAlternatingRowColors(True)
        self.setSelectionMode(QtWidgets.QAbstractItemView.NoSelection)
        # For some reason this does not work here.
        # self.setColumnCount(2)
        # header = self.horizontalHeader()
        # header.setSectionResizeMode(0,
        #                             QtWidgets.QHeaderView.ResizeToContents)
        # header.setSectionResizeMode(1, QtWidgets.QHeaderView.Stretch)

    def set_key_vals(self, keys, vals, max_key_len=23):
        """Convenience function for setting key-value pairs in this table"""
        # disable updates
        self.setUpdatesEnabled(False)
        # rows'n'cols
        self.setRowCount(len(keys))
        self.setColumnCount(2)
        # populate
        for ii, (hi, vi) in enumerate(zip(keys, vals)):
            # name
            if len(hi) < max_key_len:
                name_vis = hi
            else:
                name_vis = hi[:max_key_len-3] + "..."
            # pad with spaces (b/c css padding caused overlap)
            name_vis = " " + name_vis + " "
            label_name = self.item(ii, 0)
            if label_name is None:
                label_name = QtWidgets.QTableWidgetItem(name_vis)
                label_name.setFlags(QtCore.Qt.ItemIsEnabled)
                self.setItem(ii, 0, label_name)
            else:
                if label_name.text() != name_vis:
                    label_name.setText(name_vis)
            label_name.setToolTip(hi)
            # value
            if np.isnan(vi) or np.isinf(vi):
                fmt = "{}"
            else:
                fmt = "{:.7g}"
            value_vis = fmt.format(vi)
            label_value = self.item(ii, 1)
            if label_value is None:
                label_value = QtWidgets.QTableWidgetItem(value_vis)
                label_value.setFlags(QtCore.Qt.ItemIsEnabled)
                self.setItem(ii, 1, label_value)
            else:
                label_value.setText(value_vis)
            label_value.setToolTip(hi)
        # spacing (did not work in __init__)
        header = self.horizontalHeader()
        header.setSectionResizeMode(
            0, QtWidgets.QHeaderView.ResizeToContents)
        header.setSectionResizeMode(1, QtWidgets.QHeaderView.Stretch)
        # enable updates again
        self.setUpdatesEnabled(True)



import copy
import pkg_resources

from PyQt5 import uic, QtCore, QtWidgets


class BlockMatrix(QtWidgets.QWidget):
    pipeline_changed = QtCore.pyqtSignal(dict)
    quickviewed = QtCore.pyqtSignal(int, int)

    filter_modify_clicked = QtCore.pyqtSignal(str)
    plot_modify_clicked = QtCore.pyqtSignal(str)
    slot_modify_clicked = QtCore.pyqtSignal(str)

    def __init__(self, *args, **kwargs):
        """Helper class that wraps DataMatrix and PlotMatrix"""
        super(BlockMatrix, self).__init__(*args, **kwargs)
        path_ui = pkg_resources.resource_filename(
            "shapeout2.gui.matrix", "block_matrix.ui")
        uic.loadUi(path_ui, self)

        self._old_state = {}
        # Signals
        # DataMatrix
        self.data_matrix.matrix_changed.connect(self.on_matrix_changed)
        self.data_matrix.filter_modify_clicked.connect(
            self.filter_modify_clicked)
        self.data_matrix.slot_modify_clicked.connect(self.slot_modify_clicked)
        self.data_matrix.quickviewed.connect(self.quickviewed)
        # PlotMatrix
        self.plot_matrix.matrix_changed.connect(self.on_matrix_changed)
        self.plot_matrix.plot_modify_clicked.connect(self.plot_modify_clicked)

    def __getstate__(self):
        state = self.data_matrix.__getstate__()
        statep = self.plot_matrix.__getstate__()
        state["plots"] = statep["plots"]
        for ss in statep["elements"]:
            for plot in statep["elements"][ss]:
                state["elements"][ss][plot] = statep["elements"][ss][plot]
        return state

    def __setstate__(self, state):
        # DataMatrix
        stated = copy.deepcopy(state)
        stated.pop("plots")
        for slot_state in state["slots"]:
            slot_id = slot_state["identifier"]
            for plot_state in state["plots"]:
                plot_id = plot_state["identifier"]
                stated["elements"][slot_id].pop(plot_id)
        self.data_matrix.__setstate__(stated)
        # PlotMatrix
        statep = copy.deepcopy(state)
        statep.pop("filters")
        statep.pop("filters used")
        statep.pop("slots used")
        for slot_state in state["slots"]:
            slot_id = slot_state["identifier"]
            for filt_state in state["filters"]:
                filt_id = filt_state["identifier"]
                statep["elements"][slot_id].pop(filt_id)
        self.plot_matrix.__setstate__(statep)

    def add_dataset(self, *args, **kwargs):
        self.data_matrix.add_dataset(*args, **kwargs)

    def add_filter(self, *args, **kwargs):
        self.data_matrix.add_filter(*args, **kwargs)

    def add_plot(self, *args, **kwargs):
        self.plot_matrix.add_plot(*args, **kwargs)

    def adopt_pipeline(self, pipeline_state):
        self.__setstate__(pipeline_state)

    def enable_quickview(self, view):
        self.data_matrix.enable_quickview(view)

    def get_quickview_indices(self):
        return self.data_matrix.get_quickview_indices()

    def get_widget(self, slot_id=None, filt_plot_id=None):
        """Convenience function for testing"""
        if slot_id is None and filt_plot_id is not None:
            # get a filter or a plot
            w = self.data_matrix.filter_widgets + self.plot_matrix.plot_widgets
            for wi in w:
                if wi.identifier == filt_plot_id:
                    break
            else:
                raise KeyError(
                    "Widget identifier '{}' not found!".format(filt_plot_id))
            return wi
        elif slot_id is not None and filt_plot_id is None:
            # get a slot
            for wi in self.data_matrix.dataset_widgets:
                if wi.identifier == slot_id:
                    break
            else:
                raise KeyError(
                    "Widget identifier '{}' not found!".format(filt_plot_id))
            return wi
        elif slot_id is not None and filt_plot_id is not None:
            # get a matrix element
            wd = self.data_matrix.element_widget_dict
            wp = self.plot_matrix.element_widget_dict
            fpd = wd[slot_id]
            fpp = wp[slot_id]
            if filt_plot_id in fpp:
                wi = fpp[filt_plot_id]
            elif filt_plot_id in fpd:
                wi = fpd[filt_plot_id]
            else:
                raise KeyError(
                    "Widget identifier '{}' not found!".format(filt_plot_id))
            return wi
        else:
            raise ValueError(
                "At least one of `slot_id`, `filt_plot_id` must be specified!")

    def invalidate_elements(self, invalid_dm, invalid_pm):
        for slot_id, filt_id in invalid_dm:
            em = self.data_matrix.get_matrix_element(slot_id, filt_id)
            em.active = False
            em.invalid = True
            em.update_content()
        for slot_id, plot_id in invalid_pm:
            em = self.plot_matrix.get_matrix_element(slot_id, plot_id)
            em.active = False
            em.invalid = True
            em.update_content()

    def on_matrix_changed(self):
        state = self.__getstate__()
        self.pipeline_changed.emit(state)

    def update(self, *args, **kwargs):
        self.scrollArea_block.update()
        super(BlockMatrix, self).update(*args, **kwargs)
